from pathlib import Path
import json
from ._arcgis_model import ArcGISModel, _EmptyData
import traceback
from .._utils.env import raise_fastai_import_error
import logging

logger = logging.getLogger()

try:
    from ._image_captioning_utils import (
        image_captioner_learner,
        predict_image,
        get_bleu,
    )
    from .._utils.image_captioning_data import show_results
    from ._arcgis_model import _resnet_family
    from .._utils.common import _get_emd_path
    from ._codetemplate import image_captioning_prf

    HAS_FASTAI = True
except ImportError:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


class ImageCaptioner(ArcGISModel):

    """
    Creates an Image Captioning model.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object
                            from :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used
                            for creating the encoder of the :class:`~arcgis.learn.ImageCaptioner` ,
                            which is `resnet34` by default. It supports
                            the ResNet family of backbones.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    decoder_params          Optional dictionary. The keys of the dictionary are
                            `embed_size`, `hidden_size`, `attention_size`,
                            `teacher_forcing`, `dropout` and
                            `pretrained_embeddings`.

                            Default values:

                                | decoder_params={
                                |                     'embed_size':100,
                                |                     'hidden_size':100,
                                |                     'attention_size':100,
                                |                     'teacher_forcing':1,
                                |                     'dropout':0.1,
                                |                     'pretrained_emb':False
                                |                 }

                            Parameter Explanation:

                            - 'embed_size': Size of embedding to be used during training.
                            - 'hidden_size': Size of hidden layer.
                            - 'attention_size': Size of intermediate attention layer.
                            - 'teacher_forcing': Probability of teacher forcing.
                            - 'dropout': Dropout probability.
                            - 'pretrained_emb': If true, it will use fasttext embeddings.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.ImageCaptioner`  Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None, **kwargs):
        if not HAS_FASTAI:
            raise_fastai_import_error(
                import_exception=import_exception, message="", installation_steps=" "
            )

        super().__init__(data, backbone, **kwargs)

        if pretrained_path is not None:
            pretrained_backbone = False
        else:
            pretrained_backbone = True

        self._code = image_captioning_prf
        self.decoder_params = kwargs.get("decoder_params", {})
        self.learn = image_captioner_learner(
            self._data,
            self._backbone,
            decoder_params=self.decoder_params,
            metrics=kwargs.get("metrics", None),
            pretrained=pretrained_backbone,
        )
        if pretrained_path is not None:
            self.load(pretrained_path)  # Load model and vocab

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy", "corpus_bleu"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ImageCaptioner model from an Esri Model Definition (EMD)
        file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Optional fastai Databunch. Returned
                                data object from :meth:`~arcgis.learn.prepare_data` function or
                                None for inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.ImageCaptioner`  Object
        """

        from fastai.text.transform import Vocab

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        if data is None:
            data = _EmptyData(
                path=emd_path.parent, loss_func=None, c=0, chip_size=emd["ImageHeight"]
            )

            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)

            vocab_path = emd_path.parent / "vocab"
            data.vocab = Vocab.load(vocab_path)  # load vocab.

        return cls(data, **model_params, pretrained_path=str(model_file))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @property
    def supported_backbones(self):
        """Supported torchvision backbones for this model."""
        return ImageCaptioner._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return ImageCaptioner._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["ImageCaptioning"]

    @property
    def _model_metrics(self):
        return {"accuracy": self._get_model_metrics()}

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        if len(self.learn.recorder.metrics) == 0:
            return 0.0

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            val_losses = self.learn.recorder.val_losses
            model_accuracy = self.learn.recorder.metrics[
                self.learn._best_epoch  # index using best epoch.
            ][0]

        return float(model_accuracy)

    def bleu_score(self, **kwargs):
        """
        Computes bleu score over validation set.

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        beam_width              Optional int. The size of beam to be used
                                during beam search decoding. Default is 5.
        ---------------------   -------------------------------------------
        max_len                 Optional int. The maximum length of the
                                sentence to be decoded. Default is 20.
        =====================   ===========================================

        """
        if isinstance(self._data, _EmptyData):
            scores = self._data.emd.get("Metrics")
            if scores is None:
                logger.error("Metric not found in the loaded model")
                return
            else:
                return json.loads(scores)
        return get_bleu(self, self._data, *kwargs)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {"DataAttributes": {}, "ModelParameters": {}}
        # arcgis.learn.models._inferencing
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        # object classifier config can be used.
        _emd_template["ModelConfiguration"] = "_image_captioner_inference"
        # handle for different types of spectrums
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["ModelType"] = "ImageCaptioner"
        # Inference function of object classifier.
        _emd_template["InferenceFunction"] = "ArcGISObjectClassifier.py"

        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageCaptioner.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageCaptioner.py"

        # add encoder parameters
        _emd_template["ModelParameters"]["decoder_params"] = self.decoder_params
        # chip size
        _emd_template["DataAttributes"]["chip_size"] = self._data.chip_size
        # normalization stats
        norm_stats = []
        for k in self._data.norm_stats:
            norm_stats.append(k.tolist())
        _emd_template["DataAttributes"]["norm_stats"] = list(norm_stats)

        _emd_template["CropSizeFixed"] = 1
        _emd_template["BlackenAroundFeature"] = 0
        _emd_template["SingleLabelFieldFound"] = "Caption"

        return _emd_template

    def predict(self, path, visualize=True, **kwargs):
        return predict_image(self, path, visualize=visualize, **kwargs)

    def show_results(self, rows=4, **kwargs):
        """
        Shows the ground truth and predictions of model side by side.

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        beam_width              Optional int. The size of beam to be used
                                during beam search decoding. Default is 5.
        ---------------------   -------------------------------------------
        max_len                 Optional int. The maximum length of the
                                sentence to be decoded. Default is 20.
        =====================   ===========================================

        """
        self._check_requisites()
        return_fig = kwargs.get("return_fig", False)
        fig = show_results(self, rows=rows, **kwargs)
        if return_fig:
            return fig

    def _save(
        self, name_or_path, framework="PyTorch", publish=False, gis=None, **kwargs
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name and creates
                                all the intermediate directories.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model.
                                (Only supported by :class:`~arcgis.learn.SingleShotDetector`.)
                                If framework used is ``TF-ONNX``,
                                ``batch_size`` can be passed as an optional
                                keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an
                                item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the
                                item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise.
                                Default is False.
        =====================   ===========================================
        """
        from ._arcgis_model import _create_zip

        zip_files = kwargs.pop("zip_files", True)
        path = super()._save(
            name_or_path,
            framework=framework,
            publish=publish,
            gis=gis,
            zip_files=False,
            **kwargs
        )
        self._data.vocab.save(path / "vocab")
        if zip_files:
            _create_zip(path.name, str(path))
        return path

    def load(self, name_or_path):
        """
        Loads a compatible saved model for inferencing or fine tuning from the disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name or Path to
                                Deep Learning Package (DLPK) or
                                Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        from fastai.text.transform import Vocab

        super().load(name_or_path)
        model_path = self.learn.path / "models"
        path_like = "/" in name_or_path or "\\" in name_or_path
        name = Path(name_or_path).name if path_like else name_or_path
        if path_like:
            model_path = Path(name_or_path)
            if model_path.is_file():
                vocab_path = model_path.parent / "vocab"
            else:
                vocab_path = model_path / "vocab"
        else:
            vocab_path = model_path / name / "vocab"
        self._data.vocab = Vocab.load(vocab_path)  # load vocab.



from collections import abc
from dataclasses import dataclass, field
from typing import ClassVar

from arcgis.realtime.velocity._reserved_fields import _ReservedFields
from .time import _START_TIME_TAG, _END_TIME_TAG

_TRACK_ID_TAG: str = "TRACK_ID"


@dataclass
class _FeedTemplate:
    """
    Base Class for Feeds
    """

    # Non default fields or fields that will not be initialized by constructor
    label: str
    description: str

    # abstract properties - to be assigned by all derived classes
    _name: ClassVar[str]

    _fields: dict = field(default_factory=dict, init=False)

    # ---> Inheriting classes MUST declare the following Optional fields. <---
    # These are commented variables because of a limitation in dataclass where base class properties get ordered before
    # the derived class properties. Since these are optional properties, the init will order them before the non-optional
    # derived class properties leading to errors like - TypeError: non-default argument 'rss_url' follows default argument
    #
    # track_id_field: Optional[str] = None

    # abstract methods to be implemented by all Feeds
    def _build(self) -> dict:
        """
        Build the feed configuration object from the current state of this object. This feed configuration can be posted to the API endpoint.
        To be implemented by concrete classes.
        For example:

        :return: Feed configuration
        """
        raise NotImplemented()

    def _generate_feed_properties(self) -> dict:
        """
        Builds part of the dictionary object that is pertinent to feed and format properties from the current state of this object.
        To be implemented by concrete classes.

        :return: Dictionary object that contains feed and format properties which can be used to POST request to velocity.
        """
        raise NotImplemented

    def _generate_schema_transformation(self) -> dict:
        """
        Builds the final Schema-transformation dictionary object from the current state of _fields object in the format that can be used
        to create POST request to ArcGIS Velocity.

        :return: Dictionary object that contains the schema-transformation properties.
        """
        # validate feature_schema
        if self._fields is None or not self._fields["attributes"]:
            raise ValueError("Invalid feed schema. Cannot proceed")

        input_schema = {}
        field_mappings = []

        input_schema = self._fields.copy()
        for field in input_schema["attributes"]:
            if field["toField"]:
                if _ReservedFields.is_reserved(field["toField"]):
                    # A toField cannot be one of the reserved names.
                    raise ValueError(
                        f"'{field['toField']}' is a reserved field name. It must be renamed or dropped"
                    )

                field_mappings.append(
                    {
                        "fromField": field["name"],
                        "toField": field["toField"],
                        "tags": field["tags"],
                    }
                )

            # input_schema's attributes should not have "toField" property
            del field["toField"]

        return {
            "schemaTransformation": {
                "inputSchema": input_schema,
                "fieldMappings": field_mappings,
            }
        }

    # Feature schema manipulation methods
    def _set_fields(self, feature_schema):
        """
        Reads the feature schema config from the sample_messages response json/dict into self._fields object property.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        feature_schema         dict. The sample_messages_response["featureSchema"] json/dict object.
        ==================     ====================================================================
        """
        self._fields = feature_schema

        # validate feature_schema
        if self._fields is None or not self._fields["attributes"]:
            raise ValueError("Invalid feed schema. Cannot proceed")

        for field in self._fields["attributes"]:
            # add a property 'toField' to each attribute. This property will be used to generate the fieldMappings dict.
            field["toField"] = field["name"]

    def rename_field(self, current_name: str, new_name: str) -> bool:
        """
        Rename a field.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        current_name           String. Current field name.
        ------------------     --------------------------------------------------------------------
        new_name               String. New field name.
        ==================     ====================================================================

        :return: Field collection after transformation
        """
        if not new_name.strip():
            raise ValueError("new_name cannot be empty")
        elif _ReservedFields.is_reserved(new_name):
            raise ValueError(
                f"'{new_name}' is a reserved field name and cannot be used."
            )

        if self._fields is not None and self._fields["attributes"]:
            attributes = self._fields["attributes"]
            is_success = False
            for attr in attributes:
                if attr["name"] == current_name:
                    attr["toField"] = new_name
                    is_success = True
            if is_success:
                return True
            else:
                raise ValueError(f"could not find field `{current_name}` in the schema")
        else:
            raise Exception("Invalid feed schema. Cannot proceed")

    def change_field_data_type(self, name, new_data_type) -> bool:
        """
        Used to specify the expected data type of a field.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   String. Field name.
        ------------------     --------------------------------------------------------------------
        new_data_type          String. New data type for the field.
        ==================     ====================================================================

        :return: Boolean - True if data type change was successful
        """
        is_success = False
        if self._fields is not None and self._fields["attributes"]:
            fields = self._fields["attributes"]
            for field in fields:
                if field["name"] == name or field["toField"] == name:
                    field["dataType"] = new_data_type
                    is_success = True

            if is_success:
                return True
            else:
                raise ValueError(f"could not find field `{name}` in the schema")
        else:
            raise Exception("Invalid feed schema. Cannot proceed")

    def remove_field(self, name: str) -> bool:
        """
        Remove a field from the schema.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   String. Field to be removed from the schema.
        ==================     ====================================================================

        :return: Boolean - True if a field is removed, False otherwise
        """
        is_success = False
        if self._fields is not None and self._fields["attributes"]:
            fields = self._fields["attributes"]
            for field in fields:
                if field["name"] == name:
                    # flag the field for removal from schema by clearing the `toField` value
                    field["toField"] = None
                    if _TRACK_ID_TAG in field["tags"]:
                        field["tags"].clear()
                    elif any(
                        elem in [_START_TIME_TAG, _END_TIME_TAG]
                        for elem in field["tags"]
                    ):
                        self.reset_time_config()

                    is_success = True

        if is_success:
            return True
        else:
            raise ValueError(f"could not find field `{name}` in the schema")

    def set_track_id(self, field_name: str):
        """
         Set the track ID field for the feed.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        field_name             String. Name of the track ID field. Either the original name or renamed
                               field name can be used to specify the track ID.
        ==================     ====================================================================
        """
        is_success = False
        for field in self._fields["attributes"]:
            if field["name"] == field_name or field["toField"] == field_name:
                field["tags"] = [_TRACK_ID_TAG]
                is_success = True
            elif _TRACK_ID_TAG in field["tags"]:
                field["tags"].clear()

        if is_success:
            if self.track_id_field is None:
                self.track_id_field = field_name
            return True
        else:
            raise ValueError(f"invalid field_name: '{field_name}'")

    def _dict_deep_merge(self, dct, merge_dct):
        """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
        updating only top-level keys, dict_merge recurses down into dicts nested
        to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
        ``dct``.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        dct                    String. Dict onto which the merge is executed.
        ------------------     --------------------------------------------------------------------
        merge_dct              String. This dict will be merged into dct.
        ==================     ====================================================================
        """
        # Future enhancement - Move this to a util?
        # Or use https://anaconda.org/conda-forge/deepmerge package instead
        for k, v in merge_dct.items():
            if (
                k in dct
                and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], abc.Mapping)
            ):
                self._dict_deep_merge(dct[k], merge_dct[k])
            elif k in dct:
                raise ValueError(
                    f"property name collision found for `{k}` key. Found in dct and merge_dct dictionaries"
                )
            else:
                dct[k] = merge_dct[k]



from typing import Optional, Dict, Union, List

from arcgis import GIS

from ._bigdata_analytics import BigDataAnalytics
from ._util import _Util
import logging

_LOGGER = logging.getLogger(__name__)


class BigDataAnalyticsManager:
    """
    Used to manage big data analytic items.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    URL of the ArcGIS Velocity organization.
    ------------------     --------------------------------------------------------------------
    gis                    An authenticated :class:`arcigs.gis.GIS` object.
    ==================     ====================================================================

    """

    _gis = None
    _util = None

    def __init__(self, url: str, gis: GIS):
        self._gis = gis

        self._util = _Util(gis, url)

    @property
    def items(self) -> List[BigDataAnalytics]:
        """
        Get all big data analytic items.

        :return: returns a collection of all configured Big Data Analytics items

        .. code-block:: python

            # Get all big data analytics

            all_bigdata_analytics = bigdata_analytics.items
            all_bigdata_analytics
        """
        all_bigdata_analytics_response = self._util._get_request("analytics/bigdata")
        if (
            all_bigdata_analytics_response is not None
            and type(all_bigdata_analytics_response) is list
        ):
            bigdata_analytics_items = [
                BigDataAnalytics(self._gis, self._util, bigdata_item)
                for bigdata_item in all_bigdata_analytics_response
            ]
            return bigdata_analytics_items
        elif all_bigdata_analytics_response is None:
            _LOGGER.warning("No Big-data Analytic items found for the user.")
            return []
        else:
            raise Exception(
                f"Error retrieving Big-data Analytic items. Velocity response: ${all_bigdata_analytics_response}"
            )

    def get(self, id) -> BigDataAnalytics:
        """
        Get big data analytic items by ID.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Unique ID of a big data analytic task.
        ===============     ====================================================================

        :return: endpoint response of Big Data Analytics for the given id and label

        .. code-block:: python

            # Get big data analytics by id
            # Method: <item>.get(id)

            sample_bigdata_task = bigdata_analytics.get("id")

        """
        bigdata_analytics_item = self._util._get("analytics/bigdata", id)
        return BigDataAnalytics(self._gis, self._util, bigdata_analytics_item)



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="CascadeEncoderDecoder",
    num_stages=2,
    pretrained="open-mmlab://resnet101_v1c",
    backbone=dict(
        type="ResNetV1c",
        depth=101,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        dilations=(1, 1, 1, 1),
        strides=(1, 2, 2, 2),
        norm_cfg=norm_cfg,
        norm_eval=False,
        style="pytorch",
        contract_dilation=True,
    ),
    neck=dict(
        type="FPN", in_channels=[256, 512, 1024, 2048], out_channels=256, num_outs=4
    ),
    decode_head=[
        dict(
            type="FPNHead",
            in_channels=[256, 256, 256, 256],
            in_index=[0, 1, 2, 3],
            feature_strides=[4, 8, 16, 32],
            channels=128,
            dropout_ratio=-1,
            num_classes=19,
            norm_cfg=norm_cfg,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
            ),
        ),
        dict(
            type="PointHead",
            in_channels=[256],
            in_index=[0],
            channels=256,
            num_fcs=3,
            coarse_pred_each_layer=True,
            dropout_ratio=-1,
            num_classes=19,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
            ),
        ),
    ],
    # model training and testing settings
    train_cfg=dict(num_points=2048, oversample_ratio=3, importance_sample_ratio=0.75),
    test_cfg=dict(
        mode="whole", subdivision_steps=2, subdivision_num_points=8196, scale_factor=2
    ),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/point_rend/pointrend_r101_512x1024_80k_cityscapes/pointrend_r101_512x1024_80k_cityscapes_20200711_170850-d0ca84be.pth"



import concurrent.futures
import uuid
import datetime
import logging

_log = logging.getLogger()


def _execute(fn, **kwargs):
    """runs a generic function with optional inputs"""
    if len(kwargs) > 0:
        fn(**kwargs)
    else:
        fn()


class Job(object):
    """represents an asynchronous job"""

    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        future,
        task_name,
        jobid=None,
        task_url=None,
        notify=False,
        gis=None,
    ):
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)

        self._end_time = None
        self._url = task_url
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self._task_name} job {self._jobid}>"

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self._task_name} job {self._jobid}>"

    # ----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :return: boolean
        """
        return self._future.cancelled()

    # ----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :return: boolean
        """
        return self._future.running()

    # ----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :return: boolean
        """
        return self._future.done()

    # ----------------------------------------------------------------------
    @property
    def ellapse_time(self):
        """
        Returns the Ellapse Time for the Job
        """
        if self._end_time:
            return self._end_time - self._start_time
        else:
            return datetime.datetime.now() - self._start_time

    # ----------------------------------------------------------------------
    def _set_end_time(self, future):
        """sets the finish time"""
        self._end_time = datetime.datetime.now()

    # ----------------------------------------------------------------------
    def _notify(self, future):
        """prints finished method"""
        jobid = self._jobid
        try:
            res = future.result()
            infomsg = "{jobid} finished successfully.".format(jobid=jobid)
            _log.info(infomsg)
            print(infomsg)
        except Exception as e:
            msg = str(e)
            msg = "{jobid} failed: {msg}".format(jobid=jobid, msg=msg)
            _log.info(msg)
            print(msg)

    # ----------------------------------------------------------------------
    def result(self):
        """returns the job result"""
        return self._future.result()


class GeometryJob(Job):
    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None
    _wkid = None

    def __init__(
        self,
        future: concurrent.futures.Future,
        task_name: str,
        jobid: str = None,
        task_url: str = None,
        notify: bool = False,
        gis=None,
        out_wkid: int = None,
    ) -> None:
        super(GeometryJob, self)
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        self._wkid = out_wkid
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)

        self._end_time = None
        self._url = task_url
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid

    # ----------------------------------------------------------------------
    def result(self):
        """returns the job result"""
        from arcgis.geometry import Geometry

        if self._task_name == "lengths":
            res = self._future.result()
            return res.get("lengths", [])
        elif self._wkid:  # set the output sr to that wkid integer
            sr = {"spatialReference": {"wkid": self._wkid}}
            res = self._future.result()
            if isinstance(res, (list, tuple)):
                [g.update(sr) for g in res if not "spatialReference" in g]
                return [Geometry(g) for g in res]
            elif isinstance(res, dict) and not "spatialReference" in res:
                res.update(sr)
                return Geometry(res)
            else:
                return res
        else:
            return self._future.result()


class ItemStatusJob(Job):
    """represents an Item"""

    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        item,
        task_name,
        jobid=None,
        job_type=None,
        notify=False,
        gis=None,
    ):
        executor = concurrent.futures.ThreadPoolExecutor(1)
        future = executor.submit(self._status, *(item, jobid, job_type))
        executor.shutdown(False)
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)

        self._end_time = None
        self._url = None
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self._task_name} job>"

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self._task_name} job>"

    # ----------------------------------------------------------------------
    def _status(self, item, job_id, job_type):
        """Processes the asynchronous job"""
        import time
        from arcgis.gis import Item

        while item.status(job_id=job_id, job_type=job_type).get("status") in [
            "processing"
        ]:
            status = item.status(job_id=job_id, job_type=job_type).get("status")
            if status in ["failed", "completed"]:
                return item.status(job_id=job_id, job_type=job_type)
            else:
                return item
        return item

    # ----------------------------------------------------------------------



"""
These tools help answer one of the most common questions posed in spatial analysis: What is near what?

```create_buffers()``` creates areas of a specified distance from features.
"""
from __future__ import annotations
import json as _json

import logging as _logging
from typing import Any, Optional, Union
import arcgis as _arcgis
from arcgis.features import FeatureSet, FeatureCollection
from arcgis.features.layer import FeatureLayer, FeatureLayerCollection
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.gis import GIS, Item
from ._util import (
    _id_generator,
    _feature_input,
    _set_context,
    _create_output_service,
    GAJob,
    _prevent_bds_item,
)
from arcgis._impl.common._utils import inspect_function_inputs
from arcgis.geoprocessing import import_toolbox

_log = _logging.getLogger(__name__)

_use_async = True


def group_by_proximity(
    input_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    spatial_relationship: str,
    spatial_near_distance: Optional[float] = None,
    spatial_near_distance_unit: Optional[str] = None,
    temporal_relationship: Optional[str] = None,
    temporal_near_distance: Optional[float] = None,
    temporal_near_distance_unit: Optional[str] = None,
    output_name: Optional[str] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The Group By Proximity tool groups features that are within spatial
    proximity or spatiotemporal proximity of each other.

    ===================================================================    =============================================================================
    **Parameter**                                                           **Description**
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    input_layer                                                            Required layer. The point, line, or polygon features to be grouped.
                                                                           See :ref:`Feature Input<gaxFeatureInput>`.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    spatial_relationship                                                   Required String. The type of relationship to group by.

                                                                           Choice List:

                                                                             * ``Intersects``
                                                                             * ``Touches``
                                                                             * ``NearGeodesic``
                                                                             * ``NearPlanar``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    spatial_near_distance                                                  Optional Float. A float value used for the search distance to determine if
                                                                           features are near one another. This is only applied if ``NearGeodesic`` or
                                                                           ``NewPlanar`` is the selected ``spatial_relationship``.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    spatial_near_distance_unit                                             Optional String. The linear unit to be used with the distance value specified
                                                                           in `spatial_near_distance`. The default value is Meters.

                                                                           Choice List:

                                                                             * ``Meters``
                                                                             * ``Kilometers``
                                                                             * ``Feet``
                                                                             * ``Miles``
                                                                             * ``NauticalMiles``
                                                                             * ``Yards``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    temporal_relationship                                                  Optional String. The type of temporal relationship to group by.

                                                                           Choice List:

                                                                             * ``Intersects``
                                                                             * ``Near``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    temporal_near_distance                                                 Optional Float. A float value used for the temporal search distance to determine if features are near one another.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    temporal_near_distance_unit                                            Optional String. The temporal unit to be used with the distance value specified in `temporal_near_distance`.

                                                                           Choice List:

                                                                             * ``Milliseconds``
                                                                             * ``Seconds``
                                                                             * ``Minutes``
                                                                             * ``Hours``
                                                                             * ``Days``
                                                                             * ``Weeks``
                                                                             * ``Months``
                                                                             * ``Years``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    output_name                                                            Optional string. The task will create a feature service of the results. You define the name of the service.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    gis                                                                    Optional :class:`~arcgis.gis.GIS`. The GIS object where the analysis will take place.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    context                                                                Optional string. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                                            * ``extent`` - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                            * ``processSR`` - The features will be projected into this coordinate system for analysis.
                                                                            * ``outSR`` - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                            * ``dataStore`` - Results will be saved to the specified data store. For ArcGIS Enterprise, the default is the spatiotemporal big data store.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    future                                                                 Optional Boolean. If ``True``, a GAJob is returned instead of results. The GAJob can be queried on the status of the execution.
    ===================================================================    =============================================================================

    :return:
        * :class:`~arcgis.gis.Item` when `future=False`
        * ``GAJob`` when `future=True`

    """
    input_features = _prevent_bds_item(input_layer)

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    tbx = import_toolbox(url, gis=gis)

    if output_name is None:
        output_service_name = _id_generator(prefix="Group_By_Proximity_")
        output_name = output_service_name.replace(" ", "_")
    else:
        output_service_name = output_name.replace(" ", "_")
    if context is not None:
        output_datastore = context.get("dataStore", None)
    else:
        output_datastore = None
    output_service = _create_output_service(
        gis,
        output_name,
        output_service_name,
        "Group By Proximity",
        output_datastore=output_datastore,
    )

    params = {
        "input_features": input_features,
        "spatial_relationship": spatial_relationship,
        "spatial_near_distance": spatial_near_distance,
        "spatial_near_distance_unit": spatial_near_distance_unit,
        "temporal_relationship": temporal_relationship,
        "temporal_near_distance": temporal_near_distance,
        "temporal_near_distance_unit": temporal_near_distance_unit,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "future": True,
    }

    if output_service:
        params["output_name"] = _json.dumps(
            {
                "serviceProperties": {
                    "name": output_name,
                    "serviceUrl": output_service.url,
                },
                "itemProperties": {"itemId": output_service.itemid},
            }
        )
    else:
        params["output_name"] = output_name
        output_service = f"Results were written to: '{params['context']['dataStore']}' with the name: '{output_name}'"

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    kwargs = {}
    for key, value in params.items():
        if key != "field":
            if value is not None:
                kwargs[key] = value
        elif key == "field" and value:
            kwargs[key] = value
    params = inspect_function_inputs(tbx.group_by_proximity, **kwargs)
    params["future"] = True

    try:
        gpjob = tbx.group_by_proximity(**params)
        if future:
            return GAJob(gpjob=gpjob, return_service=output_service)
        gpjob.result()
        return output_service
    except Exception as e:
        output_service.delete()
        raise
    return


def trace_proximity_events(
    input_points: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    spatial_search_distance: float,
    spatial_search_distance_unit: str,
    temporal_search_distance: float,
    temporal_search_distance_unit: str,
    entity_id_field: str,
    entities_of_interest_ids: Optional[list[str]] = None,
    entities_of_interest_layer: Optional[
        Union[
            Item,
            FeatureCollection,
            FeatureLayer,
            FeatureLayerCollection,
            str,
            dict[str, Any],
        ]
    ] = None,
    distance_method: str = "Planar",
    include_tracks_layer: bool = False,
    max_trace_depth: Optional[int] = None,
    attribute_match_criteria: Optional[str] = None,
    output_name: Optional[str] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    The Trace Proximity Events task analyzes time-enabled point features representing moving entities.
    The task will follow entities of interest in space (location) and time to see which other entities
    the entities of interest have interacted with. The trace will continue from entity to entity to a
    configurable maximum degrees of separation from the original entity of interest.


    ===================================================================    =============================================================================
    **Parameter**                                                                                    **Description**
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    input_points                                                           Required Layer. A layer that will be used in analysis.
                                                                           See :ref:`Feature Input<gaxFeatureInput>`.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    spatial_search_distance                                                Required Float. The maximum distance between two points to be considered in
                                                                           proximity. Features closer together in space and that also meet
                                                                           `temporal_search_distance` criteria are considered in proximity of each other.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    spatial_search_distance_unit                                           Required String. The unit of of measure for `spatial_search_distance`.

                                                                           Choice List:

                                                                             * ``Meters``
                                                                             * ``Kilometers``
                                                                             * ``Feet``
                                                                             * ``Miles``
                                                                             * ``NauticalMiles``
                                                                             * ``Yards``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    temporal_search_distance                                               Required Float. The maximum duration between two points that are considered
                                                                           in proximity. Features closer together in time and that also meet the
                                                                           ``spatial_search_distance`` criteria are considered in proximity of each other.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    temporal_search_distance_unit                                          Required String. The unit of `temporal_search_distance`.

                                                                           Choice List:

                                                                             * ``Milliseconds``
                                                                             * ``Seconds``
                                                                             * ``Minutes``
                                                                             * ``Hours``
                                                                             * ``Days``
                                                                             * ``Weeks``
                                                                             * ``Months``
                                                                             * ``Years``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    entity_id_field                                                        Required String. The field used to identify distinct entities.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    entities_of_interest_ids                                               Optional List. JSON used to specify one or more entities that you are
                                                                           interested in tracing from. You can optionally include a time to start tracing
                                                                           from. If you do not specify a time, January 1, 1970, at 12:00 a.m. will be used.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    entities_of_interest_layer                                             Optional Layer. A feature class used to specify one or more entities that you
                                                                           are interested in tracing from.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    distance_method                                                        Required String. The distance type that will be used for the ``spatial_search_distance``.
                                                                           The default is ``Planar``.

                                                                           Choice List:

                                                                             * ``Planar``
                                                                             * ``Geodesic``
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    include_tracks_layer                                                   Optional Boolean. Determines whether or not an additional layer will be
                                                                           created containing the first trace event in tracks and all subsequent
                                                                           features. The default is ``False``.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    max_trace_depth                                                        Optional Integer. The maximum degrees of separation between an entity of
                                                                           interest and an entity further down the trace.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    attribute_match_criteria                                               Optional String. One or more attributes used to constrain the proximity
                                                                           events. Entities will only be considered near when the `spatial_search_distance`
                                                                           and `temporal_search_distance` criteria are met and the two entities have
                                                                           equal values of the attributes specified.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    output_name                                                            Optional string. The task will create a feature service of the results. You define the name of the service.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    gis                                                                    Optional :class:`~arcgis.gis.GIS`. The GIS object where the analysis will take place.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    context                                                                Optional string. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                                             * ``extent`` - a bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                             * ``processSR`` - The features will be projected into this coordinate system for analysis.
                                                                             * ``outSR`` - the features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                             * ``dataStore`` - Results will be saved to the specified data store. For ArcGIS Enterprise, the default is the spatiotemporal big data store.
    -------------------------------------------------------------------    -----------------------------------------------------------------------------
    future                                                                 optional Boolean. If ``True``, a GPJob is returned instead of results.
                                                                           The GPJob can be queried on the status of the execution.
    ===================================================================    =============================================================================

    :return:

        * :class:`~arcgis.gis.Item` when `future=False`
        * ``GAJob`` when `future=True`

    """
    input_points = _prevent_bds_item(input_points)

    if (
        isinstance(input_points, FeatureCollection)
        and "layers" in input_points.properties
        and len(input_points.properties.layers) > 0
    ):
        input_points = FeatureSet.from_dict(
            featureset_dict=input_points._lazy_properties.layers[0].featureSet
        )
    if (
        isinstance(entities_of_interest_layer, FeatureCollection)
        and "layers" in entities_of_interest_layer.properties
        and len(entities_of_interest_layer.layers) > 0
    ):
        entities_of_interest_layer = FeatureSet.from_dict(
            featureset_dict=entities_of_interest_layer._lazy_properties.layers[
                0
            ].featureSet
        )

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    tbx = import_toolbox(url_or_item=url, gis=gis)

    if output_name is None:
        output_service_name = _id_generator(prefix="Trace Proximity Events_")
        output_name = output_service_name.replace(" ", "_")
    else:
        output_service_name = output_name.replace(" ", "_")
    if context is not None:
        output_datastore = context.get("dataStore", None)
    else:
        output_datastore = None
    output_service = _create_output_service(
        gis,
        output_name,
        output_service_name,
        "Trace Proximity Events",
        output_datastore=output_datastore,
    )

    params = {
        "input_points": input_points,
        "entity_id_field": entity_id_field,
        "entities_of_interest_ids": entities_of_interest_ids,
        "entities_of_interest_layer": entities_of_interest_layer or "",
        "distance_method": distance_method or "Planar",
        "spatial_search_distance": spatial_search_distance,
        "spatial_search_distance_unit": spatial_search_distance_unit,
        "temporal_search_distance": temporal_search_distance,
        "temporal_search_distance_unit": temporal_search_distance_unit,
        "include_tracks_layer": include_tracks_layer,
        "max_trace_depth": max_trace_depth,
        "attribute_match_criteria": attribute_match_criteria,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "future": True,
    }

    if output_service:
        params["output_name"] = _json.dumps(
            {
                "serviceProperties": {
                    "name": output_name,
                    "serviceUrl": output_service.url,
                },
                "itemProperties": {"itemId": output_service.itemid},
            }
        )
    else:
        params["output_name"] = output_name
        output_service = f"Results were written to: '{params['context']['dataStore']}' with the name: '{output_name}'"

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    kwargs = {}
    for key, value in params.items():
        if key != "field":
            if value is not None:
                kwargs[key] = value
        elif key == "field" and value:
            kwargs[key] = value
    params = inspect_function_inputs(tbx.trace_proximity_events, **kwargs)
    params["future"] = True

    try:
        gpjob = tbx.trace_proximity_events(**params)
        if future:
            return GAJob(gpjob=gpjob, return_service=output_service)
        gpjob.result()
        return output_service
    except:
        output_service.delete()
        raise

    return


def create_buffers(
    input_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    distance: Optional[float] = None,
    distance_unit: Optional[str] = None,
    field: Optional[str] = None,
    method: str = "Planar",
    dissolve_option: str = "None",
    dissolve_fields: Optional[str] = None,
    summary_fields: Optional[list[dict[str, Any]]] = None,
    multipart: bool = False,
    output_name: Optional[str] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """

    .. image:: _static/images/create_buffers_geo/create_buffers_geo.png

    Buffers are typically used to create areas that can be further analyzed
    using other tools such as ``aggregate_points``. For example, ask the question,
    "What buildings are within one mile of the school?" The answer can be found
    by creating a one-mile buffer around the school and overlaying the buffer
    with the layer containing building footprints. The end result is a layer
    of those buildings within one mile of the school.

    ================================================    =========================================================
    **Parameter**                                       **Description**
    ------------------------------------------------    ---------------------------------------------------------
    input_layer                                         Required layer. The point, line, or polygon features to be buffered.
                                                        See :ref:`Feature Input<gaxFeatureInput>`.
    ------------------------------------------------    ---------------------------------------------------------
    distance (Required if field is not provided)        Optional float. A float value used to buffer the input features.
                                                        You must supply a value for either the distance or field parameter.
                                                        You can only enter a single distance value. The units of the
                                                        distance value are supplied by the ``distance_unit`` parameter.

                                                        The default value is 1 when `field` is None.
    ------------------------------------------------    ---------------------------------------------------------
    distance_unit (Required if distance is used)        Optional string. The linear unit to be used with the value specified in distance.

                                                        Choice list:

                                                          * ``Feet``
                                                          * ``Yards``
                                                          * ``Miles``
                                                          * ``Meters``
                                                          * ``Kilometers``
                                                          * ``NauticalMiles``

                                                        The default value is ``Miles`` when `field` is None.
    ------------------------------------------------    ---------------------------------------------------------
    field (Required if distance not provided)           Optional string. A field on the ``input_layer`` containing a buffer distance or a field expression.
                                                        A buffer expression must begin with an equal sign (=). To learn more about buffer expressions
                                                        see: `Buffer Expressions <https://developers.arcgis.com/rest/services-reference/bufferexpressions.htm>`_
    ------------------------------------------------    ---------------------------------------------------------
    method                                              Optional string. The method used to apply the buffer with. There are two methods to choose from:

                                                        Choice list:['Geodesic', 'Planar']

                                                        * ``Planar`` - This method applies a Euclidean buffers and is appropriate for local analysis on projected data. This is the default.
                                                        * ``Geodesic`` - This method is appropriate for large areas and any geographic coordinate system.
    ------------------------------------------------    ---------------------------------------------------------
    dissolve_option                                     Optional string. Determines how output polygon attributes are processed.

                                                        Choice list:

                                                          * ``All``
                                                          * ``List``
                                                          * ``None``

                                                        +--------------------------------------+---------------------------------------------------------------------------------------------------+
                                                        |Value                                 | Description                                                                                       |
                                                        +--------------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | ``All`` - All features are dissolved | You can calculate summary statistics and determine if you want multipart or single part features. |
                                                        | into one feature.                    |                                                                                                   |
                                                        +--------------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | ``List`` - Features with the same    | You can calculate summary statistics and determine if you want multipart or single part features. |
                                                        | value in the specified field         |                                                                                                   |
                                                        | will be dissolve together.           |                                                                                                   |
                                                        +--------------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | ``None`` - No features are dissolved.| There are no additional dissolve options.                                                         |
                                                        +--------------------------------------+---------------------------------------------------------------------------------------------------+
    ------------------------------------------------    ---------------------------------------------------------
    dissolve_fields                                     Specifies the fields to dissolve on. Multiple fields may be provided.
    ------------------------------------------------    ---------------------------------------------------------
    summary_fields                                      Optional string. A list of field names and statistical summary types
                                                        that you want to calculate for resulting polygons. Summary statistics
                                                        are only available if dissolveOption = List or All. By default, all
                                                        statistics are returned.

                                                        Example: ``[{"statisticType": "statistic type", "onStatisticField": "field name"}]``

                                                          * ``onStatisticField`` is the name of the field in the input point layer.

                                                          * ``statisticType`` is one of the following for numeric fields:

                                                            * ``Count`` - Totals the number of values of all the points in each polygon.
                                                            * ``Sum`` - Adds the total value of all the points in each polygon.
                                                            * ``Mean`` - Calculates the average of all the points in each polygon.
                                                            * ``Min`` - Finds the smallest value of all the points in each polygon.
                                                            * ``Max`` - Finds the largest value of all the points in each polygon.
                                                            * ``Range`` - Finds the difference between the Min and Max values.
                                                            * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                                                            * ``Var`` - Finds the variance of all the points in each polygon.

                                                          * ``statisticType`` is one of the following for string fields:

                                                            * ``Count`` - Totals the number of strings for all the points in each polygon.
                                                            * ``Any`` - Returns a sample string of a point in each polygon.

    ------------------------------------------------    ---------------------------------------------------------
    multipart                                           Optional boolean. Determines if output features are multipart or single part.
                                                        This option is only available if a ``dissolve_option`` is applied.
    ------------------------------------------------    ---------------------------------------------------------
    output_name                                         Optional string. The task will create a feature service of the results. You define the name of the service.
    ------------------------------------------------    ---------------------------------------------------------
    gis                                                 Optional, the :class:`~arcgis.gis.GIS` on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------------------    ---------------------------------------------------------
    context                                             Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                          * ``extent`` - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                          * ``processSR`` - The features will be projected into this coordinate system for analysis.
                                                          * ``outSR`` - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                          * ``dataStore`` - Results will be saved to the specified data store. For ArcGIS Enterprise, the default is the spatiotemporal big data store.
    ------------------------------------------------    ---------------------------------------------------------
    future                                              Optional boolean. If ``True``, the value is returned as a GPJob.

                                                        The default value is ``False``
    ================================================    =========================================================

    :return:
        :class:`~arcgis.features.FeatureLayerCollection`

    .. code-block:: python

            # Usage Example: To create buffer based on distance field.

            buffer = create_buffers(input_layer=lyr,
                                    field='dist',
                                    method='Geodesic',
                                    dissolve_option='All',
                                    dissolve_fields='Date')
    """
    if field and distance:
        raise ValueError(
            "Both a distance and a field cannot be specified.  Choose one and resubmit."
        )
    if field is None and distance is None:
        distance = 1
    if distance and distance_unit is None:
        distance_unit = "Miles"

    input_layer = _prevent_bds_item(input_layer)

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    if (
        isinstance(input_layer, FeatureCollection)
        and "layers" in input_layer.properties
        and len(input_layer.properties.layers) > 0
    ):
        input_layer = FeatureSet.from_dict(
            featureset_dict=input_layer._lazy_properties.layers[0].featureSet
        )
    kwargs = {
        "input_layer": input_layer,
        "distance": distance,
        "distance_unit": distance_unit,
        "field": field,
        "method": method,
        "dissolve_option": dissolve_option,
        "dissolve_fields": dissolve_fields,
        "summary_fields": summary_fields,
        "multipart": multipart,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "future": True,
    }
    params = {}
    for key, value in kwargs.items():
        if key != "field":
            if value is not None:
                params[key] = value
        elif key == "field" and value:
            params[key] = value
    if distance is None:
        params["distance"] = None
    if distance_unit is None:
        params["distance_unit"] = None
    if output_name is None:
        output_service_name = _id_generator(prefix="Create Buffers_")
        output_name = output_service_name.replace(" ", "_")
    else:
        output_service_name = output_name.replace(" ", "_")
    if context is not None:
        output_datastore = context.get("dataStore", None)
    else:
        output_datastore = None
    output_service = _create_output_service(
        gis,
        output_name,
        output_service_name,
        "Create Buffers",
        output_datastore=output_datastore,
    )

    if output_service:
        params["output_name"] = _json.dumps(
            {
                "serviceProperties": {
                    "name": output_name,
                    "serviceUrl": output_service.url,
                },
                "itemProperties": {"itemId": output_service.itemid},
            }
        )
    else:
        params["output_name"] = output_name
        output_service = f"Results were written to: '{params['context']['dataStore']}' with the name: '{output_name}'"

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    tbx = import_toolbox(url_or_item=url, gis=gis)
    params = inspect_function_inputs(tbx.create_buffers, **params)
    params["future"] = True

    try:
        gpjob = tbx.create_buffers(**params)
        if future:
            return GAJob(gpjob=gpjob, return_service=output_service)
        gpjob.result()
        return output_service
    except:
        output_service.delete()
        raise



import logging as _logging
from typing import Optional
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.gis import GIS
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis._impl.common._utils import _validate_url
from ._routing_utils import _create_toolbox

_log = _logging.getLogger(__name__)

_use_async = True

default_facilities = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
    ],
    "geometryType": "esriGeometryPoint",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_trim = {"distance": 100, "units": "esriMeters"}

default_tolerance = {"distance": 10, "units": "esriMeters"}

default_point_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {"alias": "BarrierType", "name": "BarrierType", "type": "esriFieldTypeInteger"},
        {
            "alias": "AdditionalCost",
            "name": "AdditionalCost",
            "type": "esriFieldTypeDouble",
        },
    ],
    "geometryType": "esriGeometryPoint",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_line_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "SHAPE_Length",
            "name": "SHAPE_Length",
            "type": "esriFieldTypeDouble",
        },
    ],
    "geometryType": "esriGeometryPolyline",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_polygon_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {"alias": "BarrierType", "name": "BarrierType", "type": "esriFieldTypeInteger"},
        {
            "alias": "ScaledCostFactor",
            "name": "ScaledCostFactor",
            "type": "esriFieldTypeDouble",
        },
        {
            "alias": "Shape_Length",
            "name": "Shape_Length",
            "type": "esriFieldTypeDouble",
        },
        {"alias": "Shape_Area", "name": "Shape_Area", "type": "esriFieldTypeDouble"},
    ],
    "geometryType": "esriGeometryPolygon",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_restrictions = """['Avoid Unpaved Roads', 'Avoid Private Roads', 'Driving an Automobile', 'Through Traffic Prohibited', 'Roads Under Construction Prohibited', 'Avoid Gates', 'Avoid Express Lanes', 'Avoid Carpool Roads']"""

default_attributes = {
    "fields": [
        {"alias": "ObjectID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {
            "alias": "AttributeName",
            "name": "AttributeName",
            "type": "esriFieldTypeString",
            "length": 255,
        },
        {
            "alias": "ParameterName",
            "name": "ParameterName",
            "type": "esriFieldTypeString",
            "length": 255,
        },
        {
            "alias": "ParameterValue",
            "name": "ParameterValue",
            "type": "esriFieldTypeString",
            "length": 25,
        },
    ],
    "features": [
        {
            "attributes": {
                "OBJECTID": 1,
                "AttributeName": "Any Hazmat Prohibited",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 2,
                "AttributeName": "Avoid Carpool Roads",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 3,
                "AttributeName": "Avoid Express Lanes",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 4,
                "AttributeName": "Avoid Ferries",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 5,
                "AttributeName": "Avoid Gates",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 6,
                "AttributeName": "Avoid Limited Access Roads",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 7,
                "AttributeName": "Avoid Private Roads",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 8,
                "AttributeName": "Avoid Roads Unsuitable for Pedestrians",
                "ParameterValue": "AVOID_HIGH",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 9,
                "AttributeName": "Avoid Stairways",
                "ParameterValue": "AVOID_HIGH",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 10,
                "AttributeName": "Avoid Toll Roads",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 11,
                "AttributeName": "Avoid Toll Roads for Trucks",
                "ParameterValue": "AVOID_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 12,
                "AttributeName": "Avoid Truck Restricted Roads",
                "ParameterValue": "AVOID_HIGH",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 13,
                "AttributeName": "Avoid Unpaved Roads",
                "ParameterValue": "AVOID_HIGH",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 14,
                "AttributeName": "Axle Count Restriction",
                "ParameterValue": "0",
                "ParameterName": "Number of Axles",
            }
        },
        {
            "attributes": {
                "OBJECTID": 15,
                "AttributeName": "Axle Count Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 16,
                "AttributeName": "Driving a Bus",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 17,
                "AttributeName": "Driving a Delivery Vehicle",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 18,
                "AttributeName": "Driving a Taxi",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 19,
                "AttributeName": "Driving a Truck",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 20,
                "AttributeName": "Driving an Automobile",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 21,
                "AttributeName": "Driving an Emergency Vehicle",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 22,
                "AttributeName": "Height Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 23,
                "AttributeName": "Height Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Height (meters)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 24,
                "AttributeName": "Kingpin to Rear Axle Length Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 25,
                "AttributeName": "Kingpin to Rear Axle Length Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Kingpin to Rear Axle Length (meters)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 26,
                "AttributeName": "Length Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 27,
                "AttributeName": "Length Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Length (meters)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 28,
                "AttributeName": "Preferred for Pedestrians",
                "ParameterValue": "PREFER_LOW",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 29,
                "AttributeName": "Riding a Motorcycle",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 30,
                "AttributeName": "Roads Under Construction Prohibited",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 31,
                "AttributeName": "Semi or Tractor with One or More Trailers Prohibited",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 32,
                "AttributeName": "Single Axle Vehicles Prohibited",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 33,
                "AttributeName": "Tandem Axle Vehicles Prohibited",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 34,
                "AttributeName": "Through Traffic Prohibited",
                "ParameterValue": "AVOID_HIGH",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 35,
                "AttributeName": "Truck with Trailers Restriction",
                "ParameterValue": "0",
                "ParameterName": "Number of Trailers on Truck",
            }
        },
        {
            "attributes": {
                "OBJECTID": 36,
                "AttributeName": "Truck with Trailers Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 37,
                "AttributeName": "Use Preferred Hazmat Routes",
                "ParameterValue": "PREFER_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 38,
                "AttributeName": "Use Preferred Truck Routes",
                "ParameterValue": "PREFER_MEDIUM",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 39,
                "AttributeName": "WalkTime",
                "ParameterValue": "5",
                "ParameterName": "Walking Speed (km/h)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 40,
                "AttributeName": "Walking",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 41,
                "AttributeName": "Weight Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 42,
                "AttributeName": "Weight Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Weight (kilograms)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 43,
                "AttributeName": "Weight per Axle Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 44,
                "AttributeName": "Weight per Axle Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Weight per Axle (kilograms)",
            }
        },
        {
            "attributes": {
                "OBJECTID": 45,
                "AttributeName": "Width Restriction",
                "ParameterValue": "PROHIBITED",
                "ParameterName": "Restriction Usage",
            }
        },
        {
            "attributes": {
                "OBJECTID": 46,
                "AttributeName": "Width Restriction",
                "ParameterValue": "0",
                "ParameterName": "Vehicle Width (meters)",
            }
        },
    ],
    "displayFieldName": "",
    "exceededTransferLimit": False,
}


def generate_service_areas(
    facilities: FeatureSet,
    break_values: str = "5 10 15",
    break_units: str = "Minutes",
    analysis_region: Optional[str] = None,
    travel_direction: str = "Away From Facility",
    time_of_day: Optional[datetime] = None,
    use_hierarchy: bool = False,
    uturn_at_junctions: str = "Allowed Only at Intersections and Dead Ends",
    polygons_for_multiple_facilities: str = "Overlapping",
    polygon_overlap_type: str = "Rings",
    detailed_polygons: bool = False,
    polygon_trim_distance: Optional[LinearUnit] = None,
    polygon_simplification_tolerance: Optional[LinearUnit] = None,
    point_barriers: Optional[FeatureSet] = None,
    line_barriers: Optional[FeatureSet] = None,
    polygon_barriers: Optional[FeatureSet] = None,
    restrictions: Optional[str] = None,
    attribute_parameter_values: Optional[FeatureSet] = None,
    time_zone_for_time_of_day: str = "Geographically Local",
    travel_mode: str = "Custom",
    impedance: str = "Drive Time",
    save_output_network_analysis_layer: bool = False,
    overrides: Optional[dict] = None,
    time_impedance: Optional[str] = None,
    distance_impedance: Optional[str] = None,
    polygon_detail: Optional[str] = None,
    output_type: Optional[str] = None,
    output_format: Optional[str] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """
    .. image:: _static/images/generate_service_areas/generate_service_areas.png

    This ``generate_service_areas`` tool determines network service areas around facilities. A network service area is a region that
    encompasses all streets that can be accessed within a given distance or travel time from one or more facilities. For instance, the
    10-minute service area for a facility includes all the streets that can be reached within 10 minutes from that facility. Service areas
    are commonly used to visualize and measure accessibility. For example, a three-minute drive-time polygon around a grocery store can
    determine which residents are able to reach the store within three minutes and are thus more likely to shop there.

    =================================================     ========================================================================
    **Parameter**                                          **Description**
    -------------------------------------------------     ------------------------------------------------------------------------
    facilities                                            Required :class:`~arcgis.features.FeatureSet` . The facilities around which service areas are
                                                          generated. You can load up to 1,000 facilities. The facilities feature set has an
                                                          associated attribute table. The fields in the attribute table are listed and
                                                          described below:

                                                          * ``ObjectID``: The system-managed ID field.
                                                          * ``Name``:  The name of the facility. If the name is not specified, a name is automatically
                                                            generated at solve time.

                                                          All fields from the input facilities are included in the output
                                                          polygons when the Polygons for Multiple Facilities parameter is set to Overlapping or Not
                                                          Overlapping. The ObjectID field on the input facilities is transferred to the ``FacilityOID`` field
                                                          on the output polygons.
    -------------------------------------------------     ------------------------------------------------------------------------
    break_values                                          Required string. Specifies the size and number of service area
                                                          polygons to generate for each facility. The units are determined by the Break Units value.

                                                          Multiple polygon breaks can be set to create concentric service areas per facility. For
                                                          instance, to find 2-, 3-, and 5-mile service areas for each facility, type 2 3 5, separating
                                                          the values with a space, and set Break Units to Miles. There is no limit to the number of
                                                          break values you specify.

                                                          The size of the maximum break value can't exceed the equivalent of 300 minutes or 300 miles
                                                          (482.80 kilometers). When generating detailed polygons, the maximum service-area size is
                                                          limited to 15 minutes and 15 miles (24.14 kilometers).
    -------------------------------------------------     ------------------------------------------------------------------------
    break_units                                           Required string. The unit for the Break Values parameter. The units
                                                          you choose for this parameter determine whether the tool will create service areas by
                                                          measuring driving distance or driving time. Choose a time unit to measure driving time. To
                                                          measure driving distance, choose a distance unit. Your choice also determines in which units
                                                          the tool will report total driving time or distance in the results. The choices are: Meters
                                                          Kilometers Feet Yards Miles Nautical Miles Seconds Minutes Hours Days

                                                          Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'Nautical Miles', 'Seconds',
                                                          'Minutes', 'Hours', 'Days']
    -------------------------------------------------     ------------------------------------------------------------------------
    analysis_region                                       Optional string. Specify the region in which to perform the
                                                          analysis. If a value is not specified for this parameter, the tool will automatically
                                                          calculate the region name based on the location of the input points. Setting the name of the
                                                          region is recommended to speed up the tool execution.

                                                          Choice list: ['NorthAmerica', 'SouthAmerica', 'Europe', 'MiddleEastAndAfrica', 'India',
                                                          'SouthAsia', 'SouthEastAsia', 'Thailand', 'Taiwan', 'Japan', 'Oceania', 'Greece', 'Korea']
    -------------------------------------------------     ------------------------------------------------------------------------
    travel_direction                                      Optional string. Specifies whether the direction of travel used to
                                                          generate the service area polygons is toward or away from the facilities.

                                                          * ``Away From Facility`` - The service area is generated in the direction away from the facilities.

                                                          * ``Towards Facility`` - The service area is created in the direction towards the facilities.

                                                          The direction of travel may change the shape of the polygons because impedances on opposite
                                                          sides of streets may differ or one-way restrictions may exist, such as one-way streets. The
                                                          direction you should choose depends on the nature of your service area analysis. The service
                                                          area for a pizza delivery store, for example, should be created away from the facility,
                                                          whereas the service area of a hospital should be created toward the facility.

                                                          Choice list: ['Away From Facility', 'Towards Facility']
    -------------------------------------------------     ------------------------------------------------------------------------
    time_of_day                                           Optional datetime. The time to depart from or arrive at the
                                                          facilities. The interpretation of this value depends on whether travel is toward or away from
                                                          the facilities.
                                                          It represents the departure time if Travel Direction is set to Away from Facility.
                                                          It represents the arrival time if Travel Direction is set to Toward Facility.

                                                          You can use the Time Zone for Time of Day parameter to specify whether this time and date
                                                          refers to UTC or the time zone in which the facility is located.

                                                          Repeatedly solving the same analysis, but using different Time of Day values, allows you to see
                                                          how a facility's reach changes over time. For instance, the five-minute service area around a
                                                          fire station may start out large in the early morning, diminish during the morning rush hour,
                                                          grow in the late morning, and so on, throughout the day.
    -------------------------------------------------     ------------------------------------------------------------------------
    use_hierarchy                                         Optional boolean. Specify whether hierarchy should be used when
                                                          finding the best route between the facility and the incident.

                                                          Checked (True) - Use the hierarchy attribute for the analysis. Using a hierarchy results in the
                                                          solver preferring higher-order edges to lower-order edges. Hierarchical solves are faster, and
                                                          they can be used to simulate the preference of a driver who chooses to travel on freeways over
                                                          local roads when possible-even if that means a longer trip.

                                                          Unchecked (False) - Do not use the hierarchy attribute for the analysis. Not using a hierarchy
                                                          yields an accurate service area measured along all edges of the network dataset regardless of
                                                          hierarchy level.

                                                          Regardless of whether the Use Hierarchy parameter is checked (True), hierarchy is always used when
                                                          the largest break value exceeds 240 minutes or 240 miles (386.24 kilometers).
    -------------------------------------------------     ------------------------------------------------------------------------
    uturn_at_junctions                                    Optional string. Use this parameter to restrict or permit the service
                                                          area to make U-turns at junctions. In order to understand the parameter values, consider for a
                                                          moment the following terminology: a junction is a point where a street segment ends and potentially
                                                          connects to one or more other segments; a pseudo-junction is a point where exactly two streets
                                                          connect to one another; an intersection is a point where three or more streets connect; and a
                                                          dead-end is where one street segment ends without connecting to another. Given this information, the
                                                          parameter can have the following values:

                                                          ========================================  ================================================
                                                          **Parameter**                             **Description**
                                                          ----------------------------------------  ------------------------------------------------
                                                          ALLOW_UTURNS                              |ALLOW_UTURNS|
                                                                                                    U-turns are permitted everywhere. Allowing U-turns implies
                                                                                                    that the vehicle can turn around at a junction or intersection
                                                                                                    and double back on the same street.
                                                          ----------------------------------------  ------------------------------------------------
                                                          ALLOW_DEAD_ENDS_AND                       |ALLOW_DEAD_ENDS_AND_INTERSECTIONS_ONLY|
                                                          _INTERSECTIONS_ONLY                       U-turns are prohibited at junctions where exactly two
                                                                                                    adjacent streets meet.
                                                          ----------------------------------------  ------------------------------------------------
                                                          ALLOW_DEAD_ENDS_ONLY                      |ALLOW_DEAD_ENDS_ONLY|
                                                                                                    U-turns are prohibited at all junctions and interesections
                                                                                                    and are permitted only at dead ends.
                                                          ----------------------------------------  ------------------------------------------------
                                                          NO_UTURNS                                 U-turns are prohibited at all junctions, intersections, and dead-ends.
                                                                                                    Note that even when this parameter value is chosen, a route can still
                                                                                                    make U-turns at stops. If you wish to prohibit U-turns at a stop, you can set
                                                                                                    its CurbApproach property to the appropriate value (3).

                                                                                                    The default value for this parameter is 'ALLOW_UTURNS'.
                                                          ========================================  ================================================

                                                          Choice list: ['Allowed', 'Not Allowed', 'Allowed Only at Dead Ends',
                                                          'Allowed Only at Intersections and Dead Ends']
    -------------------------------------------------     ------------------------------------------------------------------------
    polygons_for_multiple_facilities                      Optional string.  Choose how service area polygons are generated when multiple facilities are present in the analysis.

                                                          * ``Overlapping`` - Creates individual polygons for each facility. The polygons can overlap each other.
                                                            This is the default value.

                                                          * ``Not Overlapping`` - Creates individual polygons such that a polygon from one facility cannot
                                                            overlap polygons from other facilities; furthermore, any portion of the network can only be
                                                            covered by the service area of the nearest facility.
                                                          * ``Merge by Break Value`` - Creates and joins the polygons of different facilities that have the same
                                                            break value.

                                                          When using Overlapping or Not Overlapping, all fields from the input facilities are included in the
                                                          output polygons, with the exception that values from the input ObjectID field are transferred to the
                                                          FacilityOID field of the output polygons. The FacilityOID field is null when merging by break value,
                                                          and the input fields are not included in the output.

                                                          Choice list: ['Overlapping', 'Not Overlapping', 'Merge by Break Value']
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_overlap_type                                  Optional string. Specifies the option to create concentric service area
                                                          polygons as disks or rings. This option is applicable only when multiple break
                                                          values are specified for the facilities.

                                                          * ``Rings`` - The polygons representing larger breaks exclude the polygons of smaller breaks.
                                                            This creates polygons going between consecutive breaks. Use this option if you want to
                                                            find the area from one break to another. For instance, if you create 5- and 10-minute
                                                            service areas, then the 10-minute service area polygon will exclude the area under the
                                                            5-minute service area polygon. This is the default value.

                                                          * ``Disks`` - Creates polygons going from the facility to the break. For instance, if you
                                                            create 5- and 10-minute service areas, then the 10-minute service area polygon will
                                                            include the area under the 5-minute service area polygon.

                                                          Choice list: ['Rings', 'Disks']
    -------------------------------------------------     ------------------------------------------------------------------------
    detailed_polygons                                     Optional boolean. Specifies the option to create detailed or generalized
                                                          polygons.

                                                          Unchecked (False) - Creates generalized polygons, which are
                                                          generated quickly and are fairly accurate. This is the
                                                          default.

                                                          Checked (True) - Creates detailed polygons, which
                                                          accurately model the service area lines and may contain islands of
                                                          unreached areas. This option is much slower than generating
                                                          generalized polygons. This option isn't supported when using
                                                          hierarchy.
                                                          If your facilities are in an urban area with a grid-like street network,
                                                          the difference between generalized and detailed service areas would be
                                                          minimal. However, if your facilities are in a region containing mountain
                                                          and rural roads, the detailed service areas may present significantly
                                                          more accurate results than generalized service areas.

                                                          The tool supports generating detailed polygons only if the largest value specified in the Break Values
                                                          parameter is less than or equal to 15 minutes or 15 miles (24.14 kilometers).
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_trim_distance                                 Optional LinearUnit. Specifies the distance within which the service area
                                                          polygons are trimmed. This is useful when finding service areas in
                                                          places that have a sparse street network and you don't want the
                                                          service area to cover large areas where there are no street
                                                          features.

                                                          The default value is 100 meters. No value or a value of 0 for this parameter
                                                          specifies that the service area polygons should not be trimmed. This
                                                          parameter value is ignored when using hierarchy.
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_simplification_tolerance                      Optional LinearUnit. Specify by how much you want to
                                                          simplify the polygon geometry.

                                                          Simplification maintains critical vertices of a polygon to define its essential
                                                          shape and removes other vertices. The simplification distance you specify is the
                                                          maximum offset the simplified polygon boundaries can deviate from the original
                                                          polygon boundaries. Simplifying a polygon reduces the number of vertices and tends
                                                          to reduce drawing times.
    -------------------------------------------------     ------------------------------------------------------------------------
    point_barriers                                        Optional :class:`~arcgis.features.FeatureSet`  . Specify one or more points to act as temporary
                                                          restrictions or represent additional time or distance that may be required to travel on the
                                                          underlying streets. For example, a point barrier can be used to represent a fallen tree along a
                                                          street or time delay spent at a railroad crossing.

                                                          The tool imposes a limit of 250 points that can be added as barriers.
                                                          When specifying the point barriers, you can set properties for each one, such as its name or
                                                          barrier type, by using attributes. The point barriers can be specified with the following
                                                          attributes:

                                                          * ``Name``: The name of the barrier.
                                                          * ``BarrierType``: Specifies whether the point barrier restricts travel
                                                            completely or adds time or distance when it is crossed. The value
                                                            for this attribute is specified as one of the following
                                                            integers (use the numeric code, not the name in parentheses):

                                                            * 0 (Restriction) - Prohibits travel through the barrier. The barrier
                                                              is referred to as a restriction point barrier since it acts as a
                                                              restriction.

                                                            * 2 (Added Cost) - Traveling through the barrier increases the travel
                                                              time or distance by the amount specified in the Additional_Time or
                                                              Additional_Distance field. This barrier type is  referred to as an
                                                              added-cost point barrier.

                                                          * ``Additional_Time``: Indicates how much travel time is added when the
                                                            barrier is traversed. This field is applicable only for added-cost
                                                            barriers and only if the Break Units value is time based. This field
                                                            value must be greater than or equal to zero, and its units are the same
                                                            as those specified in the Break Units parameter.

                                                          * ``Additional_Distance``: Indicates how much distance is added when the
                                                            barrier is traversed. This field is applicable only for added-cost barriers
                                                            and only if the Break Units value is distance based. The field value
                                                            must be greater than or equal to zero, and its units are the same as those
                                                            specified in the Break Units parameter.
    -------------------------------------------------     ------------------------------------------------------------------------
    line_barriers                                         Optional :class:`~arcgis.features.FeatureSet`  . Specify one or more lines that prohibit travel
                                                          anywhere the lines intersect the streets. For example, a parade or protest that blocks traffic
                                                          across several street segments can be modeled with a line barrier. A line barrier can also
                                                          quickly fence off several roads from being traversed, thereby channeling possible routes away
                                                          from undesirable parts of the street network.

                                                          The tool imposes a limit on the number of streets you can
                                                          restrict using the Line Barriers parameter. While there is no limit on
                                                          the number of lines you can specify as line barriers, the combined
                                                          number of streets intersected by all the lines cannot exceed 500.

                                                          When specifying the line barriers, you can set a name property for each one by using the following
                                                          attribute:

                                                          * ``Name``: The name of the barrier.
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_barriers                                      Optional :class:`~arcgis.features.FeatureSet`  . Specify polygons that either completely restrict travel or
                                                          proportionately scale the time or distance required to travel on
                                                          the streets intersected by the polygons.
                                                          The service imposes a limit on the number of streets you
                                                          can restrict using the Polygon Barriers parameter. While there is
                                                          no limit on the number of polygons you can specify as the polygon
                                                          barriers, the combined number of streets intersected by all the
                                                          polygons should not exceed 2,000.
                                                          When specifying the polygon barriers, you can set properties for each one, such as its name or barrier type,
                                                          by using attributes. The polygon barriers can be specified with the following attributes:

                                                          * ``Name``: The name of the barrier.
                                                          * ``BarrierType``: Specifies whether the barrier restricts travel completely or scales the time or distance for traveling through it. The field value is specified as one of the following integers (use the numeric code, not the name in parentheses):

                                                            * 0 (Restriction) - Prohibits traveling through any part of the barrier.
                                                              The barrier is referred to as a restriction polygon barrier since it
                                                              prohibits traveling on streets intersected by the barrier. One use
                                                              of this type of barrier is to model floods covering areas of the
                                                              street that make traveling on those streets impossible.
                                                            * 1 (Scaled Cost) - Scales the time or distance required to travel the
                                                              underlying streets by a factor specified using the ScaledTimeFactor
                                                              or ScaledDistanceFactor fields. If the streets are partially
                                                              covered by the barrier, the travel time or distance is apportioned
                                                              and then scaled. For example, a factor 0.25 would mean that travel
                                                              on underlying streets is expected to be four times faster than
                                                              normal. A factor of 3.0 would mean it is expected to take three
                                                              times longer than normal to travel on underlying streets. This
                                                              barrier type is referred to as a scaled-cost polygon barrier. It
                                                              might be used to model storms that reduce travel speeds in specific
                                                              regions.

                                                            * ``ScaledTimeFactor``: This is the factor by which the travel time of the streets
                                                              intersected by the barrier is multiplied. This field is applicable
                                                              only for scaled-cost barriers and only if the measurement units are time
                                                              based. The field value must be greater than zero.
                                                            * ``ScaledDistanceFactor``: This is the factor by which the distance of the streets
                                                              intersected by the barrier is multiplied. This attribute is
                                                              applicable only for scaled-cost barriers and only if the measurement
                                                              units are distance based. The attribute value must be greater than
                                                              zero.
    -------------------------------------------------     ------------------------------------------------------------------------
    restrictions                                          Optional string. Specify which restrictions should be honored by the tool when finding the best routes
                                                          between facilities and demand points. A restriction represents a driving
                                                          preference or requirement. In most cases, restrictions cause roads
                                                          to be prohibited. For instance, using an Avoid Toll Roads restriction will result in a route that will
                                                          include toll roads only when it is absolutely required to travel on toll roads in order to visit an
                                                          incident or a facility. Height Restriction makes it possible to route around any clearances that are
                                                          lower than the height of your vehicle. If you are carrying corrosive materials on your vehicle, using
                                                          the Any Hazmat Prohibited restriction prevents hauling the materials along roads where it is marked as
                                                          illegal to do so.
                                                          Below is a list of available restrictions and a short description.
                                                          Some restrictions require an additional value to be
                                                          specified for their desired use. This value needs to be associated
                                                          with the restriction name and a specific parameter intended to work
                                                          with the restriction. You can identify such restrictions if their
                                                          names appear under the AttributeName column in the Attribute
                                                          Parameter Values parameter. The ParameterValue field should be
                                                          specified in the Attribute Parameter Values parameter for the
                                                          restriction to be correctly used when finding traversable roads.
                                                          Some restrictions are supported only in certain countries; their availability is stated by region in
                                                          the list below. Of the restrictions that have limited availability within a region, you can check whether
                                                          the restriction is available in a particular country by looking at the table in the Country List section
                                                          of the Data coverage for network analysis services web page. If a country has a value of  Yes in the
                                                          Logistics Attribute column, the restriction with select availability in the region is supported in that country.
                                                          If you specify restriction names that are not available in the country where your incidents are located,
                                                          the service ignores the invalid restrictions. The service also ignores restrictions whose Restriction Usage
                                                          parameter value is between 0 and 1 (see the Attribute Parameter Value parameter). It prohibits all restrictions
                                                          whose Restriction Usage parameter value is greater than 0.
                                                          The service supports the restriction names listed in the following table:

                                                          ========================================  ================================================
                                                          **Parameter**                             **Description**
                                                          ----------------------------------------  ------------------------------------------------
                                                          Any Hazmat Prohibited                     The results will not include roads
                                                                                                    where transporting any kind of hazardous material is
                                                                                                    prohibited.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Carpool Roads                       The results will avoid roads that are
                                                                                                    designated exclusively for carpool (high-occupancy)
                                                                                                    vehicles.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Express Lanes                       The results will avoid roads designated
                                                                                                    as express lanes.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Ferries                             The results will avoid ferries.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Gates                               The results will avoid roads where there are
                                                                                                    gates such as keyed access or guard-controlled
                                                                                                    entryways.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Limited Access Roads                The results will avoid roads
                                                                                                    that are limited access highways.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Private Roads                       The results will avoid roads that are
                                                                                                    not publicly owned and maintained.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Roads Unsuitable for Pedestrians    The result will avoid roads that are unsuitable for pedestrians.
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Stairways                           The result will avoid all stairways on a pedestrian suitable route.
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Toll Roads                          The results will avoid toll
                                                                                                    roads.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Toll Roads for Trucks               The result will avoid all toll roads for trucks
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Truck Restricted Roads              The result will avoid roads where trucks are not allowed except when making deliveries.
                                                          ----------------------------------------  ------------------------------------------------
                                                          Avoid Unpaved Roads                       The results will avoid roads that are
                                                                                                    not paved (for example, dirt, gravel, and so on).
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Axle Count Restriction                    The results will not include roads
                                                                                                    where trucks with the specified number of axles are prohibited. The
                                                                                                    number of axles can be specified using the Number of Axles
                                                                                                    restriction parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Driving a Bus                             The results will not include roads where
                                                                                                    buses are prohibited. Using this restriction will also ensure that
                                                                                                    the results will honor one-way streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Driving a Taxi                            The results will not include roads where
                                                                                                    taxis are prohibited. Using this restriction will also ensure that
                                                                                                    the results will honor one-way streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Driving a Truck                           The results will not include roads where
                                                                                                    trucks are prohibited. Using this restriction will also ensure that
                                                                                                    the results will honor one-way streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Driving an Automobile                     The results will not include roads
                                                                                                    where automobiles are prohibited. Using this restriction will also
                                                                                                    ensure that the results will honor one-way streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Driving an Emergency Vehicle              The results will not include
                                                                                                    roads where emergency vehicles are prohibited. Using this
                                                                                                    restriction will also ensure that the results will honor one-way
                                                                                                    streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Height Restriction                        The results will not include roads
                                                                                                    where the vehicle height exceeds the maximum allowed height for the
                                                                                                    road. The vehicle height can be specified using the Vehicle Height
                                                                                                    (meters) restriction parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Kingpin to Rear Axle                      The results will
                                                          Length Restriction                        not include roads where the vehicle length exceeds the maximum
                                                                                                    allowed kingpin to rear axle for all trucks on the road. The length
                                                                                                    between the vehicle kingpin and the rear axle can be specified
                                                                                                    using the Vehicle Kingpin to Rear Axle Length (meters) restriction
                                                                                                    parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Length Restriction                        The results will not include roads
                                                                                                    where the vehicle length exceeds the maximum allowed length for the
                                                                                                    road. The vehicle length can be specified using the Vehicle Length
                                                                                                    (meters) restriction parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Preferred for Pedestrians                 The result prefers paths designated for pedestrians.
                                                          ----------------------------------------  ------------------------------------------------
                                                          Riding a Motorcycle                       The results will not include roads
                                                                                                    where motorcycles are prohibited. Using this restriction will also
                                                                                                    ensure that the results will honor one-way streets.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Roads Under Construction Prohibited       The results will not
                                                                                                    include roads that are under construction.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Semi or Tractor with One                  The results will not include roads where semis or tractors with
                                                          or More Trailers Prohibited               one or more trailers are prohibited.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Single Axle Vehicles Prohibited           The results will not
                                                                                                    include roads where vehicles with single axles are
                                                                                                    prohibited.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Tandem Axle Vehicles Prohibited           The results will not
                                                                                                    include roads where vehicles with tandem axles are
                                                                                                    prohibited.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Through Traffic Prohibited                The results will not include
                                                                                                    roads where through traffic (non local) is prohibited.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Truck with Trailers Restriction           The results will not
                                                                                                    include roads where trucks with the specified number of trailers on
                                                                                                    the truck are prohibited. The number of trailers on the truck can
                                                                                                    be specified using the Number of Trailers on Truck restriction
                                                                                                    parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Use Preferred Hazmat Routes               The results will prefer roads
                                                                                                    that are designated for transporting any kind of hazardous
                                                                                                    materials.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Use Preferred Truck Routes                The results will prefer roads
                                                                                                    that are designated as truck routes, such as the roads that are
                                                                                                    part of the national network as specified by the National Surface
                                                                                                    Transportation Assistance Act in the United States, or roads that
                                                                                                    are designated as truck routes by the state or province, or roads
                                                                                                    that are preferred by the trucks when driving in an
                                                                                                    area.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Walking                                   The results will not include roads where
                                                                                                    pedestrians are prohibited.
                                                                                                    Availability: All countries
                                                          ----------------------------------------  ------------------------------------------------
                                                          Weight Restriction                        The results will not include roads
                                                                                                    where the vehicle weight exceeds the maximum allowed weight for the
                                                                                                    road. The vehicle weight can be specified using the Vehicle Weight
                                                                                                    (kilograms) restriction parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Weight per Axle Restriction               The results will not include
                                                                                                    roads where the vehicle weight per axle exceeds the maximum allowed
                                                                                                    weight per axle for the road. The vehicle weight per axle can be
                                                                                                    specified using the Vehicle Weight per Axle (kilograms) restriction
                                                                                                    parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ----------------------------------------  ------------------------------------------------
                                                          Width Restriction                         The results will not include roads where
                                                                                                    the vehicle width exceeds the maximum allowed width for the road.
                                                                                                    The vehicle width can be specified using the Vehicle Width (meters)
                                                                                                    restriction parameter.
                                                                                                    Availability: Select countries in North America and Europe
                                                          ========================================  ================================================

                                                          Choice list: ['Any Hazmat Prohibited', 'Avoid Carpool Roads', 'Avoid Express Lanes', 'Avoid Ferries',
                                                          'Avoid Gates', 'Avoid Limited Access Roads', 'Avoid Private Roads', 'Avoid Roads Unsuitable for Pedestrians',
                                                          'Avoid Stairways', 'Avoid Toll Roads', 'Avoid Toll Roads for Trucks', 'Avoid Truck Restricted Roads',
                                                          'Avoid Unpaved Roads', 'Axle Count Restriction', 'Driving a Bus', 'Driving a Delivery Vehicle', 'Driving a Taxi',
                                                          'Driving a Truck', 'Driving an Automobile', 'Driving an Emergency Vehicle', 'Height Restriction',
                                                          'Kingpin to Rear Axle Length Restriction', 'Length Restriction', 'Preferred for Pedestrians', 'Riding a Motorcycle',
                                                          'Roads Under Construction Prohibited', 'Semi or Tractor with One or More Trailers Prohibited',
                                                          'Single Axle Vehicles Prohibited', 'Tandem Axle Vehicles Prohibited', 'Through Traffic Prohibited',
                                                          'Truck with Trailers Restriction', 'Use Preferred Hazmat Routes', 'Use Preferred Truck Routes', 'Walking',
                                                          'Weight Restriction', 'Weight per Axle Restriction', 'Width Restriction']
    -------------------------------------------------     ------------------------------------------------------------------------
    attribute_parameter_values                            Optional :class:`~arcgis.features.FeatureSet`  . Specify additional values required by some
                                                          restrictions, such as the weight of a vehicle for Weight Restriction. You can also use the attribute
                                                          parameter to specify whether any restriction prohibits, avoids, or prefers travel on roads that use the
                                                          restriction. If the restriction is meant to avoid or prefer roads, you can further specify the degree
                                                          to which they are avoided or preferred using this parameter. For example, you can choose to never use
                                                          toll roads, avoid them as much as possible, or even highly prefer them.

                                                          The values you provide for this parameter are ignored unless Travel Mode is set to Custom. If you specify
                                                          the Attribute Parameter Values parameter from a  feature class, the field names on the feature class must
                                                          match the fields as described below:

                                                          * ``AttributeName``: Lists the name of the restriction.

                                                          * ``ParameterName``: Lists the name of the parameter associated with the
                                                            restriction. A restriction can have one or more ParameterName field
                                                            values based on its intended use.

                                                          * ``ParameterValue``: The value for ParameterName used by the tool
                                                            when evaluating the restriction.

                                                            Attribute Parameter Values is dependent on the Restrictions parameter. The ParameterValue field is
                                                            applicable only if the restriction name is specified as the value for the Restrictions parameter.

                                                            In Attribute Parameter Values, each restriction (listed as AttributeName) has a ParameterName field
                                                            value, Restriction Usage, that specifies whether the restriction prohibits, avoids, or prefers travel
                                                            on the roads associated with the restriction and the degree to which the roads are avoided or
                                                            preferred. The Restriction Usage ParameterName can be assigned any of the following string values or
                                                            their equivalent numeric values listed within the parentheses:

                                                            *  PROHIBITED (-1) - Travel on the roads using the restriction is completely prohibited.
                                                            *  AVOID_HIGH (5) - It is highly unlikely for the tool to include in the route the roads that are associated with the restriction.
                                                            *  AVOID_MEDIUM (2) - It is unlikely for the tool to include in the route the roads that are associated with the restriction.
                                                            *  AVOID_LOW (1.3) - It is somewhat unlikely for the tool to include in the route the roads that are associated with the restriction.
                                                            *  PREFER_LOW (0.8) - It is somewhat likely for the tool to include in the route the roads that are associated with the restriction.
                                                            *  PREFER_MEDIUM (0.5) - It is likely for the tool to include in the route the roads that are associated with the restriction.
                                                            *  PREFER_HIGH (0.2) - It is highly likely for the tool to include in the route the roads that are associated with the restriction.

                                                            In most cases, you can use the default value, PROHIBITED, for the Restriction Usage if the restriction
                                                            is dependent on a vehicle-characteristic such as vehicle height. However, in some cases, the value for
                                                            Restriction Usage depends on your routing preferences. For example, the Avoid Toll Roads restriction
                                                            has the default value of AVOID_MEDIUM for the Restriction Usage parameter. This means that when the
                                                            restriction is used, the tool will try to route around toll roads when it can. AVOID_MEDIUM also indicates
                                                            how important it is to avoid toll roads when finding the best route; it has a medium priority. Choosing
                                                            AVOID_LOW would put lower importance on avoiding tolls; choosing AVOID_HIGH instead would give it a higher
                                                            importance and thus make it more acceptable for the service to generate longer routes to avoid tolls.
                                                            Choosing PROHIBITED would entirely disallow travel on toll roads, making it impossible for a route to
                                                            travel on any portion of a toll road. Keep in mind that avoiding or prohibiting toll roads, and thus
                                                            avoiding toll payments, is the objective for some; in contrast, others prefer to drive on toll roads
                                                            because avoiding traffic is more valuable to them than the money spent on tolls. In the latter case, you
                                                            would choose PREFER_LOW, PREFER_MEDIUM, or PREFER_HIGH as the value for Restriction Usage. The higher the
                                                            preference, the farther the tool will go out of its way to travel on the roads associated with the
                                                            restriction.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_zone_for_time_of_day                             Optional string. Specifies the time zone or zones of the Time of Day
                                                          parameter.

                                                          * ``Geographically Local``: The Time of Day parameter refers to the time zone or zones in which the facilities
                                                            are located. Therefore, the start or end times of the service areas are staggered by time zone. Setting
                                                            Time of Day to 9:00 a.m., choosing geographically local for Time Zone for Time of Day, and solving causes
                                                            service areas to be generated for 9:00 a.m. Eastern Time for any facilities in the Eastern Time Zone,
                                                            9:00 a.m. Central Time for facilities in the Central Time Zone, 9:00 a.m. Mountain Time for facilities in
                                                            the Mountain Time Zone, and so on, for facilities in different time zones. If stores in a chain that span
                                                            the U.S. open at 9:00 a.m. local time, this parameter value could be chosen to find market territories at
                                                            opening time for all stores in one solve. First, the stores in the Eastern Time Zone open and a polygon is
                                                            generated, then an hour later stores open in Central Time, and so on. Nine o'clock is always in local time
                                                            but staggered in real time.

                                                          * ``UTC``: The Time of Day parameter refers to Coordinated Universal Time (UTC).
                                                            Therefore, all facilities are reached or departed from simultaneously, regardless of the time zone each is
                                                            in. Setting Time of Day to 2:00 p.m., choosing UTC, then solving causes service areas to be generated for
                                                            9:00 a.m. Eastern Standard Time for any facilities in the Eastern Time Zone, 8:00 a.m. Central Standard
                                                            Time for facilities in the Central Time Zone, 7:00 a.m. Mountain Standard Time for facilities in the
                                                            Mountain Time Zone, and so on, for facilities in different time zones. The scenario above assumes standard
                                                            time. During daylight saving time, the Eastern, Central, and Mountain Times would each be one hour ahead
                                                            (that is, 10:00, 9:00, and 8:00 a.m., respectively). One of the cases in which the UTC option is useful is
                                                            to visualize emergency-response coverage for a jurisdiction that is split into two time zones. The emergency
                                                            vehicles are loaded as facilities. Time of Day is set to now in UTC. (You need to determine what  the current
                                                            time and date are  in terms of UTC to correctly use this option.) Other properties are set and the analysis
                                                            is solved. Even though a time-zone boundary divides the vehicles, the results show areas that can be reached
                                                            given current traffic conditions. This same process  can be used for other times as well, not just for now.

                                                          Irrespective of the Time Zone for Time of Day setting, all facilities must be in the same time zone
                                                          when Time of Day has a nonnull value and Polygons for Multiple Facilities is set to create merged or
                                                          nonoverlapping polygons.

                                                          Choice list: ['Geographically Local', 'UTC']
    -------------------------------------------------     ------------------------------------------------------------------------
    travel_mode                                           Optional string.  Specify the mode of transportation to model in the analysis. Travel
                                                          modes are managed in ArcGIS Online and can be configured by the administrator of your organization to better
                                                          reflect your organization's workflows. You need to specify the name of a travel mode supported by your
                                                          organization.

                                                          To get a list of supported travel mode names, run the GetTravelModes tool from the Utilities toolbox
                                                          available under the same GIS Server connection you used to access the tool. The GetTravelModes tool
                                                          adds a table, Supported Travel Modes, to the application. Any value in the Travel Mode Name field from the
                                                          Supported Travel Modes table can be specified as input. You can also specify the value from Travel Mode
                                                          Settings field as input. This speeds up the tool execution as the tool does not have to lookup the settings
                                                          based on the travel mode name.

                                                          The default value, Custom, allows you to configure your own travel mode using the custom travel mode
                                                          parameters (UTurn at Junctions, Use Hierarchy, Restrictions, Attribute Parameter Values,  and Impedance).
                                                          The default values of the custom travel mode parameters model travelling by car. You may want to choose
                                                          Custom and set the custom travel mode parameters listed above to model a pedestrian with a fast walking speed
                                                          or a truck with a given height, weight, and cargo of certain hazardous materials. You may choose to do this
                                                          to try out different settings to get desired analysis results. Once you have identified the analysis settings,
                                                          you should work with your organization's administrator and save these settings as part of new or existing
                                                          travel mode so that everyone in your organization can rerun the analysis with the same settings.
    -------------------------------------------------     ------------------------------------------------------------------------
    impedance                                             Optional string. Specify the impedance, which is a value that represents the effort or cost
                                                          of traveling along road segments or on other parts of the transportation network.
                                                          Travel distance is an impedance; the length of a road in kilometers can be thought of as impedance. Travel
                                                          distance in this sense is the same for all modes-a kilometer for a pedestrian is also a kilometer for a car.
                                                          (What may change is the pathways on which the different modes are allowed to travel, which affects distance
                                                          between points, and this is modeled by travel mode settings.) Travel time can also be an impedance; a car
                                                          may take one minute to travel a mile along an empty road. Travel times can vary by travel mode-a pedestrian
                                                          may take more than 20  minutes to walk the same mile, so it is important to choose the right impedance for
                                                          the travel mode you are modeling.  Choose from the following impedance values: Drive Time-Models travel times
                                                          for a car. These travel times are dynamic and fluctuate according to traffic flows in areas where traffic data
                                                          is available. This is the default value. Truck Time-Models travel times for a truck.  These travel times are
                                                          static for each road and don't fluctuate with traffic. Walk Time-Models travel times for a pedestrian. Travel
                                                          Distance-Stores  length measurements along roads and paths. To model walk distance, choose this option and
                                                          ensure Walking is  set in the Restriction parameter. Similarly, to model drive or truck distance, choose Travel
                                                          Distance here and set the appropriate restrictions so your vehicle travels only on roads where it is permitted
                                                          to do so. The value you provide for this parameter is ignored unless Travel Mode is set to Custom, which is the
                                                          default value. If you choose Drive Time, Truck Time, or Walk Time, the Measurement Units parameter must be set
                                                          to a time-based value; if you choose Travel Distance for Impedance, Measurement Units must be distance-based.

                                                          Choice list:['Drive Time', 'Truck Time', 'Walk Time', 'Travel Distance']
    -------------------------------------------------     ------------------------------------------------------------------------
    save_output_network_analysis_layer                    Optional boolean. Specify if the tool should save the analysis settings as a network analysis layer file.
                                                          You cannot directly work with this file even when you open the file in an ArcGIS Desktop application like ArcMap.
                                                          It is meant to be sent to Esri Technical Support to diagnose the quality of results returned from the tool.
                                                          True: Save the network analysis layer file. The file is downloaded in a temporary directory on your machine.
                                                          In ArcGIS Pro, the location of the downloaded file can be determined by viewing the value for the Output Network Analysis
                                                          Layer parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap,
                                                          the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Network
                                                          Analysis Layer parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.
                                                          False: Do not save the network analysis layer file. This is the default.
    -------------------------------------------------     ------------------------------------------------------------------------
    overrides                                             Optional string. Specify additional settings that can influence the behavior of the solver when finding solutions
                                                          for the network analysis problems. The value for this parameter needs to be specified in dict. For example, a valid value is
                                                          of the following form {"overrideSetting1" : "value1", "overrideSetting2" :
                                                          "value2"}. The override setting name is always enclosed in double quotes. The values can be a number, Boolean,
                                                          or string. The default value for this parameter is no value, which indicates not to override any solver settings. Overrides
                                                          are advanced settings that should be used only after careful analysis of the results obtained before and after applying
                                                          the settings. A list of supported override settings for each solver and their acceptable values can be obtained by contacting
                                                          Esri Technical Support.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_impedance                                        Optional string. Specify the time-based impedance.
    -------------------------------------------------     ------------------------------------------------------------------------
    distance_impedence                                    Optional string. Specify the distance-based impedance.
    -------------------------------------------------     ------------------------------------------------------------------------
    polygon_detail                                        Optional string. Specify the detail of the polygon you want to create.

                                                          Choice list: ["Generalized", "Standard", "High"]
    -------------------------------------------------     ------------------------------------------------------------------------
    output_format                                         Optional. Specify the format in which the output features are created.

                                                          Choose from the following formats:

                                                          * Feature Set - The output features are returned as feature classes and tables. This is the default.
                                                          * JSON File - The output features are returned as a compressed file containing the
                                                            JSON representation of the outputs. When this option is specified, the output is a single
                                                            file (with a .zip extension) that contains one or more JSON files (with a .json extension)
                                                            for each of the outputs created by the service.
                                                          * GeoJSON File - The output features are returned as a compressed file containing the GeoJSON
                                                            representation of the outputs. When this option is specified, the output is a single file
                                                            (with a .zip extension) that contains one or more GeoJSON files (with a .geojson extension)
                                                            for each of the outputs created by the service.
    -------------------------------------------------     ------------------------------------------------------------------------
    gis                                                   Optional, the :class:`~arcgis.gis.GIS` on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------------------     ------------------------------------------------------------------------
    future                                                Optional boolean. If True, a future object will be returned and the process
                                                          will not wait for the task to complete. The default is False, which means wait for results.
    =================================================     ========================================================================

    :return: the following as a named tuple:

    * service_areas - Service Areas as a FeatureSet
    * solve_succeeded - Solve Succeeded as a boolean

    Click `GenerateServiceAreas <https://developers.arcgis.com/rest/network/api-reference/service-area-asynchronous-service.htm>`_ for additional help.

    .. code-block:: python

            # Usage Example: To determine network service areas around facilities at time breaks of 5, 10, 15 min of drive time.

            agg_result = generate_service_areas(facilities=facilities,
                                                break_values=[5, 10, 15],
                                                break_units="Minutes",
                                                time_of_day=current_time)
    """

    from arcgis.geoprocessing import import_toolbox

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.asyncServiceArea.url[
        : -len("/GenerateServiceAreas")
    ]
    url = _validate_url(url, gis)
    tbx = _create_toolbox(url, gis)
    defaults = dict(
        zip(
            tbx.generate_service_areas.__annotations__.keys(),
            tbx.generate_service_areas.__defaults__,
        )
    )
    if facilities is None:
        facilities = defaults["facilities"]
    if polygon_trim_distance is None:
        polygon_trim_distance = defaults["polygon_trim_distance"]
    if polygon_simplification_tolerance is None:
        polygon_simplification_tolerance = defaults["polygon_simplification_tolerance"]
    if point_barriers is None:
        point_barriers = defaults["point_barriers"]
    if line_barriers is None:
        line_barriers = defaults["line_barriers"]
    if polygon_barriers is None:
        polygon_barriers = defaults["polygon_barriers"]
    if restrictions is None:
        restrictions = defaults["restrictions"]
    if attribute_parameter_values is None:
        attribute_parameter_values = defaults["attribute_parameter_values"]
    if overrides is None:
        overrides = defaults["overrides"]
    if time_impedance is None and "time_impedance" in defaults:
        time_impedance = defaults["time_impedance"]
    if distance_impedance is None and "distance_impedance" in defaults:
        distance_impedance = defaults["distance_impedance"]
    if polygon_detail is None:
        polygon_detail = defaults.get("polygon_detail", None)
    if output_type is None:
        output_type = defaults.get("output_type", None)
    if output_format is None:
        output_format = defaults.get("output_format", None)

    if isinstance(break_values, list):
        break_values = " ".join(map(str, break_values))
    from arcgis._impl.common._utils import inspect_function_inputs

    params = {
        "facilities": facilities,
        "break_values": break_values,
        "break_units": break_units,
        "analysis_region": analysis_region,
        "travel_direction": travel_direction,
        "time_of_day": time_of_day,
        "use_hierarchy": use_hierarchy,
        "uturn_at_junctions": uturn_at_junctions,
        "polygons_for_multiple_facilities": polygons_for_multiple_facilities,
        "polygon_overlap_type": polygon_overlap_type,
        "detailed_polygons": detailed_polygons,
        "polygon_trim_distance": polygon_trim_distance,
        "polygon_simplification_tolerance": polygon_simplification_tolerance,
        "point_barriers": point_barriers,
        "line_barriers": line_barriers,
        "polygon_barriers": polygon_barriers,
        "restrictions": restrictions,
        "attribute_parameter_values": attribute_parameter_values,
        "time_zone_for_time_of_day": time_zone_for_time_of_day,
        "travel_mode": travel_mode,
        "impedance": impedance,
        "save_output_network_analysis_layer": save_output_network_analysis_layer,
        "overrides": overrides,
        "time_impedance": time_impedance,
        "distance_impedance": distance_impedance,
        "polygon_detail": polygon_detail,
        "output_type": output_type,
        "output_format": output_format,
        "gis": gis,
        "future": True,
    }
    params = inspect_function_inputs(tbx.generate_service_areas, **params)
    params["future"] = True
    job = tbx.generate_service_areas(**params)
    if future:
        return job
    return job.result()


generate_service_areas.__annotations__ = {
    "facilities": FeatureSet,
    "break_values": str,
    "break_units": str,
    "analysis_region": str,
    "travel_direction": str,
    "time_of_day": datetime,
    "use_hierarchy": bool,
    "uturn_at_junctions": str,
    "polygons_for_multiple_facilities": str,
    "polygon_overlap_type": str,
    "detailed_polygons": bool,
    "polygon_trim_distance": LinearUnit,
    "polygon_simplification_tolerance": LinearUnit,
    "point_barriers": FeatureSet,
    "line_barriers": FeatureSet,
    "polygon_barriers": FeatureSet,
    "restrictions": str,
    "attribute_parameter_values": FeatureSet,
    "time_zone_for_time_of_day": str,
    "travel_mode": str,
    "impedance": str,
    "return": tuple,
}



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
model = dict(
    neck=dict(
        type="FPN_CARAFE",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        num_outs=5,
        start_level=0,
        end_level=-1,
        norm_cfg=None,
        act_cfg=None,
        order=("conv", "norm", "act"),
        upsample_cfg=dict(
            type="carafe",
            up_kernel=5,
            up_group=1,
            encoder_kernel=3,
            encoder_dilation=1,
            compressed_channels=64,
        ),
    )
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/carafe/faster_rcnn_r50_fpn_carafe_1x_coco/faster_rcnn_r50_fpn_carafe_1x_coco_bbox_mAP-0.386_20200504_175733-385a75b7.pth"



import json
from pathlib import Path
from ._codetemplate import image_classifier_prf
from functools import partial
from ._arcgis_model import ArcGISModel
import logging

logger = logging.getLogger()

try:
    from fastai.basic_train import Learner
    from ._arcgis_model import (
        _resnet_family,
        _vgg_family,
        _densenet_family,
        _set_ddp_multigpu,
        _isnotebook,
    )
    from ._timm_utils import filter_timm_models
    from ._unet_utils import is_no_color, show_results_multispectral
    import torch
    from torch import nn
    import torch.nn.functional as F
    from torchvision import models
    from ._arcgis_model import _EmptyData, _change_tail
    from torchvision.models._utils import IntermediateLayerGetter
    from collections import OrderedDict
    import numpy as np
    from fastai.torch_core import split_model_idx
    from .._utils.classified_tiles import per_class_metrics
    from fastai.vision import flatten_model
    from .._utils.segmentation_loss_functions import FocalLoss, MixUpCallback, DiceLoss

    #
    try:
        from torchvision.models.segmentation.segmentation import _segm_model
    except:
        from torchvision.models.segmentation.segmentation import (
            _segm_resnet as _segm_model,
        )
    #
    from torchvision.models.segmentation.deeplabv3 import DeepLabHead, DeepLabV3
    from torchvision.models.segmentation.fcn import FCNHead
    from ._deeplab_utils import Deeplab, compute_miou
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._psp_utils import accuracy
    from ._PointRend import PointRendSemSegHead, PointRend_target_transform
    from .._utils.env import is_arcgispronotebook
    import matplotlib.pyplot as plt

    HAS_FASTAI = True
except Exception as e:

    class DeepLabV3:
        pass

    HAS_FASTAI = False


class _DeepLabOverride(DeepLabV3):
    """
    class to override the DeepLabV3 class such that after forwrd pass we can
    take output as a tuple instead of dictionary in parent class.
    """

    def __init__(
        self,
        chip_size,
        num_class,
        backbone,
        classifier,
        aux_classifier=None,
        pointrend=False,
        keep_dilation=False,
    ):
        super().__init__(backbone, classifier, aux_classifier)
        self.pointrend = pointrend

        if self.pointrend:
            return_layers = {"layer4": "out"}
            return_layers["layer3"] = "aux"
            return_layers["layer2"] = "res2"
            return_layers["layer1"] = "res1"
            self.backbone = IntermediateLayerGetter(
                backbone, return_layers=return_layers
            )
            subdivision_steps = 3
            stride = 8
            if not keep_dilation:
                subdivision_steps = 4
                stride = 16
                remove_dilation = list(self.backbone.children())[-2]
                change_dilation = list(self.backbone.children())[-1]
                for n, m in remove_dilation.named_modules():
                    if n == "0.conv2":
                        m.dilation, m.padding, m.stride = (1, 1), (1, 1), (2, 2)
                    elif "conv2" in n:
                        m.dilation, m.padding, m.stride = (1, 1), (1, 1), (1, 1)
                    elif "0.downsample.0" in n:
                        m.stride = (2, 2)

                for n, m in change_dilation.named_modules():
                    if "0.conv2" in n:
                        m.dilation, m.padding, m.stride = (1, 1), (1, 1), (1, 1)
                    elif "conv2" in n:
                        m.dilation, m.padding, m.stride = (2, 2), (2, 2), (1, 1)
                    elif "0.downsample.0" in n:
                        m.stride = (1, 1)

            self.pointrend_head = PointRendSemSegHead(
                num_class,
                256,
                train_num_points=(chip_size / stride) ** 2,
                subdivision_num_points=(chip_size / (stride / 2)) ** 2,
                subdivision_steps=subdivision_steps,
            )  # backbone_features_channel 256

    def forward(self, x):
        if self.pointrend:
            result = self.modified_forward(x)
            if self.training:
                return result["out"], result["aux"], result["pointrend"]
            else:
                return result["pointrend"]
        else:
            result = super().forward(x)
            if self.training:
                return result["out"], result["aux"]
            else:
                return result["out"]

    def modified_forward(self, x):
        input_shape = x.shape[-2:]
        features = self.backbone(x)
        result = OrderedDict()
        x = features["out"]
        x = self.classifier(x)
        result["pointrend"] = self.pointrend_head(x, [features["res1"]])
        result["out"] = F.interpolate(
            x, size=input_shape, mode="bilinear", align_corners=False
        )

        if self.aux_classifier is not None:
            x = features["aux"]
            x = self.aux_classifier(x)
            result["aux"] = F.interpolate(
                x, size=input_shape, mode="bilinear", align_corners=False
            )

        return result


def _create_deeplab(
    chip_size, num_class, pretrained=True, pointrend=True, keep_dilation=False, **kwargs
):
    """
    Create default torchvision pretrained model with resnet101.
    """
    # model = models.segmentation.deeplabv3_resnet101(pretrained=True, progress=True, **kwargs)

    model = _segm_model("deeplabv3", "resnet101", 21, True, pretrained_backbone=False)
    if pretrained:
        state_dict = models.utils.load_state_dict_from_url(
            models.segmentation.segmentation.model_urls["deeplabv3_resnet101_coco"]
        )
        model.load_state_dict(state_dict)
    model = _DeepLabOverride(
        chip_size,
        num_class,
        model.backbone,
        model.classifier,
        model.aux_classifier,
        pointrend,
        keep_dilation,
    )
    model.classifier = DeepLabHead(2048, num_class)
    model.aux_classifier = FCNHead(1024, num_class)

    return model


class DeepLab(ArcGISModel):
    """
    Model architecture from https://arxiv.org/abs/1706.05587.
    Creates a :class:`~arcgis.learn.DeepLab` Image Segmentation/ Pixel Classification model,
    based on https://github.com/pytorch/vision/tree/master/torchvision/models/segmentation.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone convolutional neural network
                            model used for feature extraction, which
                            is `resnet101` by default since it is pretrained in
                            torchvision.
                            Supported backbones: ResNet, DenseNet, VGG family and
                            specified Timm models(experimental support) from
                            :func:`~arcgis.learn.DeepLab.backbones`.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    pointrend               Optional boolean. If True, it will use PointRend
                            architecture on top of the segmentation head.
                            Default: False. PointRend architecture from
                            https://arxiv.org/pdf/1912.08193.pdf.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    class_balancing         Optional boolean. If True, it will balance the
                            cross-entropy loss inverse to the frequency
                            of pixels per class. Default: False.
    ---------------------   -------------------------------------------
    mixup                   Optional boolean. If True, it will use mixup
                            augmentation and mixup loss. Default: False
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. If True, it will use focal loss.
                            Default: False
    ---------------------   -------------------------------------------
    dice_loss_fraction      Optional float.
                            Min_val=0, Max_val=1
                            If > 0 , model will use a combination of default or
                            focal(if focal=True) loss with the specified fraction
                            of dice loss.
                            E.g.
                            for dice = 0.3, loss = (1-0.3)*default loss + 0.3*dice
                            Default: 0
    ---------------------   -------------------------------------------
    dice_loss_average       Optional str.

                            * micro: Micro dice coefficient will be used for loss calculation.

                            * macro: Macro dice coefficient will be used for loss calculation.

                            A macro-average will compute the metric independently
                            for each class and then take the average (hence treating
                            all classes equally), whereas a micro-average will
                            aggregate the contributions of all classes to compute the
                            average metric. In a multi-class classification setup,
                            micro-average is preferable if you suspect there might be
                            class imbalance (i.e you may have many more examples of
                            one class than of other classes)
                            Default: 'micro'
    ---------------------   -------------------------------------------
    ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
    ---------------------   -------------------------------------------
    keep_dilation           Optional boolean. When PointRend architecture is used,
                            keep_dilation=True can potentially improves accuracy
                            at the cost of memory consumption. Default: False
    =====================   ===========================================

    :return: :class:`~arcgis.learn.DeepLab` Object
    """

    def __init__(
        self,
        data,
        backbone=None,
        pretrained_path=None,
        pointrend=False,
        *args,
        **kwargs,
    ):
        # Set default backbone to be 'resnet101'
        if backbone is None:
            backbone = models.resnet101

        if pretrained_path is not None:
            pretrained_backbone = False
        else:
            pretrained_backbone = True

        self._check_dataset_support(data)
        if not (self._check_backbone_support(backbone)):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)

        self._pointrend = pointrend

        self._ignore_classes = kwargs.get("ignore_classes", [])
        if self._ignore_classes != [] and len(data.classes) <= 3:
            raise Exception(
                f"`ignore_classes` parameter can only be used when the dataset has more than 2 classes."
            )

        data_classes = list(self._data.class_mapping.keys())
        if 0 not in list(data.class_mapping.values()):
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes if k != 0
            ]
        else:
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes
            ]
        if self._ignore_classes != []:
            if 0 not in self._ignore_mapped_class:
                self._ignore_mapped_class.insert(0, 0)
            global accuracy
            accuracy = partial(accuracy, ignore_mapped_class=self._ignore_mapped_class)

        self.mixup = kwargs.get("mixup", False)
        self.class_balancing = kwargs.get("class_balancing", False)
        self.focal_loss = kwargs.get("focal_loss", False)
        self.dice_loss_fraction = kwargs.get("dice_loss_fraction", False)
        self.weighted_dice = kwargs.get("weighted_dice", False)
        self.keep_dilation = kwargs.get("keep_dilation", False)
        self.dice_loss_average = kwargs.get("dice_loss_average", "micro")

        self._code = image_classifier_prf
        if (
            self._backbone.__name__ == "resnet101"
            and "timm" not in self._backbone.__module__
        ):
            model = _create_deeplab(
                data.chip_size,
                data.c,
                pretrained=pretrained_backbone,
                pointrend=self._pointrend,
                keep_dilation=self.keep_dilation,
            )
            if self._is_multispectral:
                model = _change_tail(model, data)
        else:
            model = Deeplab(
                data.c,
                self._backbone,
                data.chip_size,
                self._pointrend,
                keep_dilation=self.keep_dilation,
                pretrained=pretrained_backbone,
            )

        if not _isnotebook():
            _set_ddp_multigpu(self)
            if self._multigpu_training:
                self.learn = Learner(data, model, metrics=accuracy).to_distributed(
                    self._rank_distributed
                )
                self._map_location = {"cuda:%d" % 0: "cuda:%d" % self._rank_distributed}
            else:
                self.learn = Learner(data, model, metrics=accuracy)
        else:
            self.learn = Learner(data, model, metrics=accuracy)

        self.learn.loss_func = self._deeplab_loss

        class_weight = None
        if self.class_balancing:
            if data.class_weight is not None:
                # Handle condition when nodata is already at pixel value 0 in data
                if (data.c - 1) == data.class_weight.shape[0]:
                    class_weight = (
                        torch.tensor(
                            [data.class_weight.mean()] + data.class_weight.tolist()
                        )
                        .float()
                        .to(self._device)
                    )
                else:
                    class_weight = (
                        torch.tensor(data.class_weight).float().to(self._device)
                    )
            else:
                if getattr(data, "overflow_encountered", False):
                    logger.warning(
                        "Overflow Encountered. Ignoring `class_balancing` parameter."
                    )
                    class_weight = [1] * len(data.classes)
                else:
                    logger.warning(
                        "Could not find 'NumPixelsPerClass' in 'esri_accumulated_stats.json'. Ignoring `class_balancing` parameter."
                    )

        if self._ignore_classes != []:
            if not self.class_balancing:
                class_weight = torch.tensor([1] * data.c).float().to(self._device)
            class_weight[self._ignore_mapped_class] = 0.0

        self._final_class_weight = class_weight

        if self.focal_loss:
            self.learn.loss_func = FocalLoss(self.learn.loss_func)
        if self.mixup:
            self.learn.callbacks.append(MixUpCallback(self.learn))
        if self.dice_loss_fraction:
            self.learn.loss_func = DiceLoss(
                self.learn.loss_func,
                self.dice_loss_fraction,
                weighted_dice=self.weighted_dice,
                dice_average=self.dice_loss_average,
            )

        self.learn.model = self.learn.model.to(self._device)
        self._freeze()
        self._arcgis_init_callback()  # make first conv weights learnable
        if pretrained_path is not None:
            self.load(pretrained_path)

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return DeepLab._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return DeepLab._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(
            [
                "*dpn*",
                "*inception*",
                "*nasnet*",
                "*repvgg*",
                "*resnetblur*",
                "*selecsls*",
                "*tresnet*",
                "*hrnet*",
            ]
        )
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))
        return [*_resnet_family, *_densenet_family, *_vgg_family] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return DeepLab._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Classified_Tiles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.DeepLab` semantic segmentation object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.DeepLab` Object
        """

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            empty_data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=emd["ImageHeight"],
            )
            empty_data.class_mapping = class_mapping
            empty_data.color_mapping = color_mapping
            empty_data._is_empty = True
            empty_data = get_multispectral_data_params_from_emd(empty_data, emd)
            empty_data.emd_path = emd_path
            empty_data.emd = emd
            return cls(empty_data, **model_params, pretrained_path=str(model_file))
        else:
            return cls(data, **model_params, pretrained_path=str(model_file))

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {"ModelParameters": {}}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_deeplab_infrencing"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        _emd_template["ModelType"] = "ImageClassification"
        _emd_template["ModelParameters"]["pointrend"] = self._pointrend
        _emd_template["ModelParameters"]["keep_dilation"] = self.keep_dilation
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["ignore_mapped_class"] = self._ignore_mapped_class
        _emd_template["SupportsVariableTileSize"] = True
        _emd_template["Classes"] = []
        class_data = {}
        for i, class_name in enumerate(
            self._data.classes[1:]
        ):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = (
                [random.choice(range(256)) for i in range(3)]
                if is_no_color(self._data.color_mapping)
                else self._data.color_mapping[inverse_class_mapping[class_name]]
            )
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    def accuracy(self):
        """Computes per pixel accuracy on validation set."""
        try:
            return self.learn.validate()[1].tolist()
        except Exception as e:
            accuracy = self._data.emd.get("accuracy")
            if accuracy:
                return accuracy
            else:
                logger.error("Metric not found in the loaded model")

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy"]

    @property
    def _model_metrics(self):
        return {"accuracy": "{0:1.4e}".format(self._get_model_metrics())}

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        if len(self.learn.recorder.metrics) == 0:
            return 0.0

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            val_losses = self.learn.recorder.val_losses
            model_accuracy = self.learn.recorder.metrics[
                val_losses.index(min(val_losses))
            ][0]
        return float(model_accuracy)

    def _deeplab_loss(self, outputs, targets, **kwargs):
        targets = targets.squeeze(1).detach()

        criterion = nn.CrossEntropyLoss(weight=self._final_class_weight).to(
            self._device
        )
        if self.learn.model.training:
            out = outputs[0]
            aux = outputs[1]
            if self._pointrend:
                pointrend_out = outputs[2][0]
                pointrend_coord = outputs[2][1]
                pointrend_target = PointRend_target_transform(targets, pointrend_coord)
        else:  # validation
            out = outputs
        main_loss = criterion(out, targets)

        if self.learn.model.training:
            aux_loss = criterion(aux, targets)
            if self._pointrend:
                pointrend_loss = criterion(pointrend_out, pointrend_target)
                total_loss = main_loss + 0.4 * aux_loss + pointrend_loss
            else:
                total_loss = main_loss + 0.4 * aux_loss
            return total_loss
        else:
            return main_loss

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (nn.BatchNorm2d)):
                continue
            if hasattr(i, "dilation"):
                dilation = i.dilation
                dilation = dilation[0] if isinstance(dilation, tuple) else dilation
                if dilation > 1:
                    break
            for p in i.parameters():
                p.requires_grad = False

        self.learn.layer_groups = split_model_idx(
            self.learn.model, [idx]
        )  ## Could also call self.learn.freeze after this line because layer groups are now present.
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self.learn.show_results(
            rows=rows, ignore_mapped_class=self._ignore_mapped_class, **kwargs
        )
        if is_arcgispronotebook():
            plt.show()

    def _show_results_multispectral(
        self, rows=5, alpha=0.7, **kwargs
    ):  # parameters adjusted in kwargs
        """
        Shows the ground truth and predictions of model side by side.

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Number of rows of data to be displayed, if
                                batch size is smaller, then the rows will
                                display the value provided for batch size.
        ---------------------   -------------------------------------------
        alpha                   Optional Float. Opacity parameter for label
                                overlay on image. Float [0.0 - 1.0]
                                Default: 0.7
        =====================   ===========================================

        """
        return_fig = kwargs.get("return_fig", False)
        ret_val = show_results_multispectral(self, nrows=rows, alpha=alpha, **kwargs)
        if return_fig:
            fig, ax = ret_val
            return fig

    def mIOU(self, mean=False, show_progress=True):
        """
        Computes mean IOU on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        num_classes = torch.arange(self._data.c)
        miou = compute_miou(
            self,
            self._data.valid_dl,
            mean,
            num_classes,
            show_progress,
            self._ignore_mapped_class,
        )

        if mean:
            miou = [
                miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class
            ]
            return np.mean(miou)
        if self._ignore_mapped_class == []:
            return dict(zip(["0"] + self._data.classes[1:], miou))
        else:
            class_values = [0] + list(self._data.class_mapping.keys())
            return {
                class_values[i]: miou[i]
                for i in range(len(miou))
                if i not in self._ignore_mapped_class
            }

    def per_class_metrics(self, ignore_classes=[]):
        """
        Computer per class precision, recall and f1-score on validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        self                    segmentation model object -> [PSPNetClassifier | UnetClassifier | DeepLab]
        ---------------------   -------------------------------------------
        ignore_classes          Optional list. It will contain the list of class
                                values on which model will not incur loss.
                                Default: []
        =====================   ===========================================

        Returns per class precision, recall and f1 scores
        """
        ignore_classes = np.unique(self._ignore_classes + ignore_classes).tolist()
        try:
            self._check_requisites()
            ## Calling imported function `per_class_metrics`
            return per_class_metrics(self, ignore_classes)
        except:
            import pandas as pd

            return pd.read_json(self._data.emd["per_class_metrics"])
